//----------------------------------------------------------------------------------
// File:   BilateralBlurEngine.cpp
// Author: Miguel Sainz & Louis Bavoil
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

#include "BilateralBlurEngine.h"
#include "ScreenSpaceAOPaths.h"

#define MAX_PATH_STR 512

#define BLUR

enum
{
    IDC_SHARPNESS_STATIC,
    IDC_SHARPNESS_SLIDER,
    IDC_BLUR_SIZE_STATIC,
    IDC_BLUR_SIZE_SLIDER,
    IDC_EDGE_THRESHOLD_STATIC,
    IDC_EDGE_THRESHOLD_SLIDER,
};

static CDXUTDialog *m_myMenu = NULL;

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
BilateralBlurEngine::BilateralBlurEngine()
{
    // Buffer sizes
    m_InWidth        = 0;
    m_InHeight       = 0;
    m_OutWidth       = 0;
    m_OutHeight      = 0;

    m_Effect         = NULL;
    m_DSStencilWrite = NULL;
    m_DSStencilTest  = NULL;

    m_TempBuffer     = NULL;
    m_EdgeBlurredAO  = NULL;
    m_EdgeStencilAO  = NULL;
    m_pSharpness     = NULL;

    m_D3DDevice		 = NULL;

    m_BlurRadius     = 7;
    m_Sharpness      = 16.0f;
    m_EdgeThreshold  = 0.1f;

    m_NumSamples     = 1;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::CreateDepthStencilStates()
{
    HRESULT hr;

    // Create the depthstencil states for the rendering pass in MSAA
    D3D10_DEPTH_STENCIL_DESC dsDesc;
    
    dsDesc.DepthEnable      = FALSE;
    dsDesc.DepthWriteMask   = D3D10_DEPTH_WRITE_MASK_ZERO;
    
    dsDesc.StencilEnable    = TRUE;
    dsDesc.StencilReadMask  = 0x0;
    dsDesc.StencilWriteMask = 0xFF;

    // Stencil operations if pixel is front-facing
    dsDesc.FrontFace.StencilFailOp      = D3D10_STENCIL_OP_REPLACE ;
    dsDesc.FrontFace.StencilDepthFailOp = D3D10_STENCIL_OP_REPLACE ;
    dsDesc.FrontFace.StencilPassOp      = D3D10_STENCIL_OP_REPLACE ;
    dsDesc.FrontFace.StencilFunc        = D3D10_COMPARISON_ALWAYS;
    dsDesc.BackFace.StencilFailOp       = D3D10_STENCIL_OP_REPLACE ;
    dsDesc.BackFace.StencilDepthFailOp  = D3D10_STENCIL_OP_REPLACE ;
    dsDesc.BackFace.StencilPassOp       = D3D10_STENCIL_OP_REPLACE ;
    dsDesc.BackFace.StencilFunc         = D3D10_COMPARISON_ALWAYS;

    SAFE_RELEASE(m_DSStencilWrite);
    V( m_D3DDevice->CreateDepthStencilState(&dsDesc, &m_DSStencilWrite) );

    dsDesc.DepthEnable      = FALSE;
    dsDesc.DepthWriteMask   = D3D10_DEPTH_WRITE_MASK_ZERO;
    
    dsDesc.StencilEnable    = TRUE;
    dsDesc.StencilReadMask  = 0xFF;
    dsDesc.StencilWriteMask = 0x0;

    // Stencil operations if pixel is front-facing
    dsDesc.FrontFace.StencilFailOp      = D3D10_STENCIL_OP_KEEP;
    dsDesc.FrontFace.StencilDepthFailOp = D3D10_STENCIL_OP_KEEP;
    dsDesc.FrontFace.StencilPassOp      = D3D10_STENCIL_OP_KEEP;
    dsDesc.FrontFace.StencilFunc        = D3D10_COMPARISON_EQUAL;
    dsDesc.BackFace.StencilFailOp       = D3D10_STENCIL_OP_KEEP;
    dsDesc.BackFace.StencilDepthFailOp  = D3D10_STENCIL_OP_KEEP;
    dsDesc.BackFace.StencilPassOp       = D3D10_STENCIL_OP_KEEP;
    dsDesc.BackFace.StencilFunc         = D3D10_COMPARISON_EQUAL;

    SAFE_RELEASE(m_DSStencilTest);
    V( m_D3DDevice->CreateDepthStencilState(&dsDesc, &m_DSStencilTest) );
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT BilateralBlurEngine::OnResizedSwapChain( int inputWidth, int inputHeight, int outputWidth, int outputHeight )
{
    HRESULT hr = S_OK;

    // Recalculate buffer sizes
    m_InWidth   = inputWidth;
    m_InHeight  = inputHeight;
    m_OutWidth  = outputWidth;
    m_OutHeight = outputHeight;
  
    // Create non-MSAA RT
    D3D10_TEXTURE2D_DESC pDesc;
    pDesc.Width              = m_OutWidth;
    pDesc.Height             = m_OutHeight;
    pDesc.MipLevels          = 1;
    pDesc.ArraySize          = 1;
    pDesc.Format             = DXGI_FORMAT_R8_UNORM;
    pDesc.SampleDesc.Count   = 1;
    pDesc.SampleDesc.Quality = 0;
    pDesc.Usage              = D3D10_USAGE_DEFAULT;
    pDesc.BindFlags          = D3D10_BIND_SHADER_RESOURCE | D3D10_BIND_RENDER_TARGET;
    pDesc.CPUAccessFlags     = 0;
    pDesc.MiscFlags          = 0;

    SAFE_DELETE(m_TempBuffer);
    m_TempBuffer = new SimpleRT( m_D3DDevice , &pDesc);

    SAFE_DELETE(m_EdgeBlurredAO);
    m_EdgeBlurredAO = new SimpleRT( m_D3DDevice , &pDesc);
  
    SAFE_DELETE(m_EdgeStencilAO);
    pDesc.Format    = DXGI_FORMAT_D24_UNORM_S8_UINT;
    pDesc.BindFlags = D3D10_BIND_DEPTH_STENCIL; // | D3D10_BIND_SHADER_RESOURCE;    
    m_EdgeStencilAO = new SimpleDS( m_D3DDevice , &pDesc);
 
    UpdateBlur();

    CreateDepthStencilStates();

    m_viewportInput.TopLeftX  = 0;
    m_viewportInput.TopLeftY  = 0;
    m_viewportInput.MinDepth  = 0;
    m_viewportInput.MaxDepth  = 1;
    m_viewportInput.Width     = m_InWidth;
    m_viewportInput.Height    = m_InHeight;

    m_viewportOutput.TopLeftX = 0;
    m_viewportOutput.TopLeftY = 0;
    m_viewportOutput.MinDepth = 0;
    m_viewportOutput.MaxDepth = 1;
    m_viewportOutput.Width    = m_OutWidth;
    m_viewportOutput.Height   = m_OutHeight;

    return hr;
}
//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::OnReleasingSwapChain( )
{
    m_InWidth  = 0;
    m_InHeight = 0;
    m_OutWidth  = 0;
    m_OutHeight = 0;
    
    SAFE_DELETE(m_TempBuffer);
    SAFE_DELETE(m_EdgeBlurredAO);
    SAFE_DELETE(m_EdgeStencilAO);
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::OnDestroyDevice( )
{
    SAFE_RELEASE(m_Effect);
    SAFE_RELEASE(m_DSStencilWrite);
    SAFE_RELEASE(m_DSStencilTest);
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT	BilateralBlurEngine::OnFrameRender( ID3D10RenderTargetView   *Destination, 
                                            ID3D10ShaderResourceView *Diffuse,
                                            ID3D10ShaderResourceView *Source,
                                            ID3D10ShaderResourceView *SourceGuidance,
                                            ID3D10ShaderResourceView *FullResDepth,
                                            ID3D10ShaderResourceView *MultiSampleDepth)
{
    HRESULT hr = S_OK;

    // Select the right technique
    ID3D10EffectTechnique* currentTechnique    = 0;
    ID3D10EffectTechnique* currentTechnique_SS = 0;
    if( Diffuse )
    {
        currentTechnique    = m_TechniqueBlurWithDiffuse;
        currentTechnique_SS = m_TechniqueBlurWithDiffuse_SS;
    }
    else
    {
        currentTechnique    = m_TechniqueBlur;
        currentTechnique_SS = m_TechniqueBlur_SS;
    }

    // Update shader variables g_Resolution and g_InvResolution
    m_Resolution[0]    = (float)m_OutWidth; 
    m_Resolution[1]    = (float)m_OutHeight; 
    m_InvResolution[0] = 1.0f / m_Resolution[0];
    m_InvResolution[1] = 1.0f / m_Resolution[1];
    m_pResolution->SetFloatVector   ( m_Resolution    );
    m_pInvResolution->SetFloatVector( m_InvResolution );

    // In MSAA mode run an edge detector on the to find which pixels should be
    // solved using supersampling.
    if( m_NumSamples > 1 ) {
        m_D3DDevice->ClearDepthStencilView( m_EdgeStencilAO->pDSV, D3D10_CLEAR_STENCIL , 1.0, 0 );
        m_D3DDevice->OMGetDepthStencilState( &m_DSDefault, 0 );

        V( m_SrcTexVar->SetResource( Source ) );
        V( m_DepTexVar->SetResource( FullResDepth ) );
        V( m_ColTexVar->SetResource( Diffuse ) );
        V( m_MSAADepTexVar->SetResource( MultiSampleDepth ) );

        m_D3DDevice->RSSetViewports(1, &m_viewportOutput );	

        // Fill Stencil
        m_D3DDevice->OMSetDepthStencilState(m_DSStencilWrite, 1);
        m_D3DDevice->OMSetRenderTargets( 0, NULL, m_EdgeStencilAO->pDSV );
        m_TechniqueEdgeDetection->GetPassByIndex(0)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );

        // Blur Pass X
        m_D3DDevice->OMSetRenderTargets( 1, &m_TempBuffer->pRTV, m_EdgeStencilAO->pDSV );
        m_D3DDevice->OMSetDepthStencilState(m_DSStencilTest, 0);
        currentTechnique->GetPassByIndex(0)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );

        m_D3DDevice->OMSetDepthStencilState(m_DSStencilTest, 1);
        currentTechnique_SS->GetPassByIndex(0)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );
        
        // Blur Pass Y
        m_D3DDevice->OMSetRenderTargets( 1, &Destination, m_EdgeStencilAO->pDSV );
        m_D3DDevice->OMSetDepthStencilState(m_DSStencilTest, 0);
        V( m_SrcTexVar->SetResource( m_TempBuffer->pSRV ) );
        currentTechnique->GetPassByIndex(1)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );

        m_D3DDevice->OMSetDepthStencilState(m_DSStencilTest, 1);
        currentTechnique_SS->GetPassByIndex(1)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );

        // Unbind SRVs to avoid warning from the fx framework
        m_SrcTexVar->SetResource( NULL );
        m_DepTexVar->SetResource( NULL );
        m_ColTexVar->SetResource( NULL );
        m_MSAADepTexVar->SetResource( NULL );
        currentTechnique->GetPassByIndex(0)->Apply( 0 );

        // Reset the depth-stencil state
        m_D3DDevice->OMSetDepthStencilState(m_DSDefault, 0);
        SAFE_RELEASE(m_DSDefault);
    }
    else
    {
        // Blur Pass X
        m_D3DDevice->RSSetViewports(1, &m_viewportOutput );	
        m_D3DDevice->OMSetRenderTargets( 1, &m_TempBuffer->pRTV, NULL );
        V( m_SrcTexVar->SetResource( Source ) );
        V( m_DepTexVar->SetResource( FullResDepth ) );
        V( m_ColTexVar->SetResource( Diffuse ) );
        currentTechnique->GetPassByIndex(0)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );
        
        // Blur Pass Y
        m_D3DDevice->RSSetViewports(1, &m_viewportOutput );
        m_D3DDevice->OMSetRenderTargets( 1, &Destination, NULL );
        V( m_SrcTexVar->SetResource( m_TempBuffer->pSRV ) );
        V( m_ColTexVar->SetResource( Diffuse ) );
        currentTechnique->GetPassByIndex(1)->Apply( 0 );
        m_D3DDevice->Draw( 3, 0 );

        // Unbind SRVs to avoid warning from the fx framework
        m_SrcTexVar->SetResource( NULL );
        m_DepTexVar->SetResource( NULL );
        m_ColTexVar->SetResource( NULL );
        m_MSAADepTexVar->SetResource( NULL );
        m_DepthBufferVar->SetResource( NULL );
        currentTechnique->GetPassByIndex(0)->Apply( 0 );
    }

    return hr;
}

HRESULT	BilateralBlurEngine::OnFrameRender( ID3D10RenderTargetView   *Destination, 
                                            ID3D10ShaderResourceView *Diffuse )
{
    HRESULT hr = S_OK;

    m_D3DDevice->RSSetViewports(1, &m_viewportOutput);
    m_D3DDevice->OMSetRenderTargets( 1, &Destination, NULL );

    if (Diffuse)
    {
        // Blur Passthrough
        V( m_ColTexVar->SetResource( Diffuse ) );
        V( m_TechniquePassthrough->GetPassByIndex(0)->Apply( 0 ) );
        m_D3DDevice->Draw( 3, 0 );
        V( m_ColTexVar->SetResource( NULL ) );
        V( m_TechniquePassthrough->GetPassByIndex(0)->Apply( 0 ) );
    }
    else
    {
        float clear_color[4] = {1.0f, 1.0f, 1.0f, 1.0f};
        m_D3DDevice->ClearRenderTargetView( Destination, clear_color);
    }

    return hr;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::CreateMenu ( CDXUTDialog *myMenu )
{
    int iY = 300;
    WCHAR sz[100];
    
    int dy = 22;
    StringCchPrintf( sz, 100, L"Blur sharpness: %0.0f", m_Sharpness ); 
    myMenu->AddStatic( IDC_SHARPNESS_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_SHARPNESS_SLIDER, 50, iY += dy, 100, 22, 0, 32, (int)m_Sharpness );

    StringCchPrintf( sz, 100, L"Blur size: %0.0f", 2*m_BlurRadius+1 );
    myMenu->AddStatic( IDC_BLUR_SIZE_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_BLUR_SIZE_SLIDER, 50, iY += dy, 100, 22, 0, 16, (int)m_BlurRadius );

    StringCchPrintf( sz, 100, L"Edge threshold: %0.2f", m_EdgeThreshold );
    myMenu->AddStatic( IDC_EDGE_THRESHOLD_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_EDGE_THRESHOLD_SLIDER, 50, iY += dy, 100, 22, 0, 100, (int)(m_EdgeThreshold * 100) );

    m_myMenu = myMenu;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::OnGUIEvent( CDXUTDialog *myMenu, UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {		
        case IDC_SHARPNESS_SLIDER: 
        {
            m_Sharpness = (float)myMenu->GetSlider( IDC_SHARPNESS_SLIDER )->GetValue();

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Blur sharpness: %0.0f", m_Sharpness ); 
            myMenu->GetStatic( IDC_SHARPNESS_STATIC )->SetText( sz );

            UpdateSharpness();
            break;
        }
        case IDC_BLUR_SIZE_SLIDER:
        {
            m_BlurRadius = (float)myMenu->GetSlider( IDC_BLUR_SIZE_SLIDER )->GetValue();

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Blur size: %0.0f", 2*m_BlurRadius+1 );
            myMenu->GetStatic( IDC_BLUR_SIZE_STATIC )->SetText( sz );

            UpdateBlur();
            break;
        }
        case IDC_EDGE_THRESHOLD_SLIDER:
        {
            m_EdgeThreshold = (float)myMenu->GetSlider( IDC_EDGE_THRESHOLD_SLIDER )->GetValue() * 0.01f;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Edge threshold: %0.2f", m_EdgeThreshold );
            myMenu->GetStatic( IDC_EDGE_THRESHOLD_STATIC )->SetText( sz );

            UpdateBlur();
            break;
        }
    }
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::UpdateBlur()
{
    float radius     = m_BlurRadius;
    float sigma      = (radius+1) / 2;
    float inv_sigma2 = 1.0f / (2*sigma*sigma);
    
    m_pBlurFalloff->SetFloat(inv_sigma2);
    m_pBlurRadius->SetFloat(radius);

    m_pEdgeThreshold->SetFloat( m_EdgeThreshold );
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
inline float sqr(float x)
{
    return x * x;
}

void BilateralBlurEngine::UpdateSharpness()
{
    float sharpness = sqr(m_Sharpness / m_SceneScale);
    if (m_pSharpness) m_pSharpness->SetFloat(sharpness);
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::SetSceneScale(float scale)
{
    m_SceneScale = scale;
    UpdateSharpness();
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT BilateralBlurEngine::OnCreateDevice( ID3D10Device* pd3dDevice )
{
    HRESULT hr = S_OK;

    m_D3DDevice = pd3dDevice;

    return hr;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::SetNumSamplesPerPixel(int numSamples)
{
    m_NumSamples = numSamples;
    RecompileShader();
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void BilateralBlurEngine::RecompileShader()
{
    ID3D10Effect *tempEffect = NULL;

    D3D10_SHADER_MACRO shaderMacros[3];
    memset(shaderMacros, 0, sizeof(shaderMacros));

    CHAR numSamplesStr[2];
    StringCchPrintfA(numSamplesStr, 3, "%d", m_NumSamples);
    shaderMacros[0].Name = "NUM_MSAA_SAMPLES";
    shaderMacros[0].Definition = numSamplesStr;

    // Create the effect
    WCHAR str[MAX_PATH_STR+1];
    DXUTFindDXSDKMediaFileCch( str, MAX_PATH_STR, DEFAULT_SHADERPATH L"\\ScreenSpaceAO\\BilateralBlurEngine.fx" );
    D3DX10CreateEffectFromFile( str, shaderMacros, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, m_D3DDevice, NULL, NULL, &tempEffect, NULL, NULL);
    if( tempEffect && tempEffect->IsValid() )
    {
        SAFE_RELEASE(m_Effect);
        m_Effect = tempEffect;
    }
    else
    {
        MessageBox(NULL, L"Effect Compilation failed.", L"ERROR", MB_OK|MB_SETFOREGROUND|MB_TOPMOST);	
        return;
    }
    
    // Obtain the techniques
    m_TechniqueBlur                          = m_Effect->GetTechniqueByName( "BlurPass" );
    m_TechniqueBlurWithDiffuse               = m_Effect->GetTechniqueByName( "BlurPassWithDiffuse" );
    m_TechniqueBlur_SS                       = m_Effect->GetTechniqueByName( "BlurPassSupersampling" );
    m_TechniqueBlurWithDiffuse_SS            = m_Effect->GetTechniqueByName( "BlurPassWithDiffuseSupersampling" );
    m_TechniquePassthrough                   = m_Effect->GetTechniqueByName( "BlurPassthrough" );
    m_TechniqueEdgeDetection                 = m_Effect->GetTechniqueByName( "BlurEdgeDetection" );

    // Obtain pointers to the shader variales
    m_ColTexVar      = m_Effect->GetVariableByName( "tColor"  )->AsShaderResource();
    m_DepTexVar      = m_Effect->GetVariableByName( "tDepth"   )->AsShaderResource();
    m_MSAADepTexVar  = m_Effect->GetVariableByName( "tMSAADepth" )->AsShaderResource();
    m_SrcTexVar      = m_Effect->GetVariableByName( "tSource" )->AsShaderResource();
    m_DepthBufferVar = m_Effect->GetVariableByName( "tDepthBuffer" )->AsShaderResource();
    
    m_pSharpness     = m_Effect->GetVariableByName( "g_Sharpness"     )->AsScalar();
    m_pEdgeThreshold = m_Effect->GetVariableByName( "g_EdgeThreshold" )->AsScalar();
    m_pBlurRadius    = m_Effect->GetVariableByName( "g_BlurRadius"    )->AsScalar();
    m_pBlurFalloff   = m_Effect->GetVariableByName( "g_BlurFalloff"   )->AsScalar();
    m_pResolution    = m_Effect->GetVariableByName( "g_Resolution"    )->AsVector();
    m_pInvResolution = m_Effect->GetVariableByName( "g_InvResolution" )->AsVector();
  
    // Set Defaults
    UpdateBlur();
    UpdateSharpness();
}
