//--------------------------------------------------------------------------------------
// File:    BilateralBlurEngine.fx
// Author:  Miguel Sainz & Louis Bavoil
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#ifndef NUM_MSAA_SAMPLES
#define NUM_MSAA_SAMPLES 2
#endif

Texture2D<float4> tColor;
Texture2D<float>  tSource;
Texture2D<float>  tDepth;
Texture2D<float>  tDepthBuffer;
Texture2DMS<float, NUM_MSAA_SAMPLES> tMSAADepth;

float2 g_Resolution;
float2 g_InvResolution;
float g_BlurRadius;
float g_BlurFalloff;
float g_Sharpness;
float g_EdgeThreshold;
float2 g_ZLinParams;

SamplerState samNearest
{
    Filter   = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState samLinear
{
    Filter   = MIN_MAG_MIP_LINEAR;
    AddressU = Clamp;
    AddressV = Clamp;
};

//--------------------------------------------------------------------------------------
DepthStencilState DisableDepth
{
    DepthEnable    = FALSE;
    DepthWriteMask = ZERO;
};

BlendState DisableBlend
{
    BlendEnable[0] = false;
};

BlendState EnableBlend
{
    BlendEnable[0] = true;
};



//--------------------------------------------------------------------------------------
struct PostProc_VSOut
{
    float4 pos : SV_Position;
    float2 tex : TEXCOORD0;
};

//Vertex shader that generates a full screen quad with texcoords g_Rom vertIDs
//To use draw 3 vertices with primitive type triangle strip
PostProc_VSOut FullScreenQuadVS( uint id : SV_VertexID )
{
    PostProc_VSOut output = (PostProc_VSOut)0.0f;

    output.tex = float2( (id << 1) & 2, id & 2 );
    output.pos = float4( output.tex * float2( 2.0f, -2.0f ) + float2( -1.0f, 1.0f) , 0.0f, 1.0f );
    
    return output;
}

//-------------------------------------------------------------------------
float fetch_eye_z(float2 uv)
{
    float z = tDepth.Sample(samNearest, float3(uv, 0));
    return z;
}

//-------------------------------------------------------------------------
float BlurFunction(float2 uv, float r, float center_c, float center_d, inout float w_total)
{
    float c = tSource.Sample( samNearest, uv );
    float d = fetch_eye_z(uv);

    float ddiff = d - center_d;
    float w = exp(-r*r*g_BlurFalloff - ddiff*ddiff*g_Sharpness);
    w_total += w;

    return w*c;
}

//-------------------------------------------------------------------------
float4 BlurX( PostProc_VSOut IN ): SV_TARGET
{
    float b = 0;
    float w_total = 0;
    float center_c = tSource.Sample( samNearest, IN.tex );
    float center_d = fetch_eye_z(IN.tex);
    
    for (float r = -g_BlurRadius; r <= g_BlurRadius; ++r)
    {
        float2 uv = IN.tex.xy + float2(r*g_InvResolution.x , 0);
        b += BlurFunction(uv, r, center_c, center_d, w_total);	
    }

    return b/w_total;
}

float4 BlurX_SS( PostProc_VSOut IN ): SV_TARGET
{
    float b = 0;
    float w_total = 0;
    float center_c = tSource.Sample( samNearest, IN.tex );

    for (float r = -g_BlurRadius; r <= g_BlurRadius; ++r) 
    {
        for (int sample = 0; sample < NUM_MSAA_SAMPLES; ++sample)
        {
            float center_d = tMSAADepth.Load( int2(IN.pos.xy), sample );
            float2 uv = IN.tex.xy + float2(r*g_InvResolution.x , 0);
            b += BlurFunction(uv, r, center_c, center_d, w_total);	
        }
    }

    return b/w_total;
}

//-------------------------------------------------------------------------
float4 BlurY( uniform bool combine, PostProc_VSOut IN): SV_TARGET
{
    float b = 0;
    float w_total = 0;
    float center_c = tSource.Sample( samNearest, IN.tex, 0 );
    float center_d = fetch_eye_z(IN.tex);
    
    for (float r = -g_BlurRadius; r <= g_BlurRadius; ++r)
    {
        float2 uv = IN.tex.xy + float2(0, r*g_InvResolution.y); 
        b += BlurFunction(uv, r, center_c, center_d, w_total);
    }
    if (combine) {
        return  b/w_total*tColor.Sample(samNearest, float3(IN.tex, 0));
    }
    return b/w_total;	
}

float4 BlurY_SS( uniform bool combine, PostProc_VSOut IN): SV_TARGET
{
    //return 0;
    float b = 0;
    float w_total = 0;
    float center_c = tSource.Sample( samNearest, IN.tex, 0 );
 
    for (float r = -g_BlurRadius; r <= g_BlurRadius; ++r)
    {
        for (int sample = 0; sample < NUM_MSAA_SAMPLES; ++sample)
        {
            float center_d = tMSAADepth.Load( int2(IN.pos.xy), sample );
            float2 uv = IN.tex.xy + float2(0, r*g_InvResolution.y);
            b += BlurFunction(uv, r, center_c, center_d, w_total);
        }
    }

    if (combine)
    {
        return  b/w_total*tColor.Sample(samNearest, float3(IN.tex, 0), 0);
    }
    return b/w_total;	
}


//-------------------------------------------------------------------------
float4 Passthrough_PS( PostProc_VSOut IN ): SV_TARGET
{
    return tColor.SampleLevel(samNearest, float3(IN.tex, 0), 0);
}

//-------------------------------------------------------------------------
technique10 BlurPass
{
    pass pX
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurX() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }

    pass pY
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurY(false) ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique10 BlurPassWithDiffuse
{
    pass pX
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurX() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }

    pass pY
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurY(true) ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique10 BlurPassSupersampling
{
    pass pX
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurX_SS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }

    pass pY
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurY_SS(false) ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

technique10 BlurPassWithDiffuseSupersampling
{
    pass pX
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurX_SS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }

    pass pY
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurY_SS(true) ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}

//-------------------------------------------------------------------------
technique10 BlurPassthrough
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, Passthrough_PS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepth, 0 );

    }
}

//--------------------------------------------------------------------------
float edgeDetectScalar(float sx, float sy, float threshold)
{
    float dist = (sx*sx+sy*sy);
    float e = (dist > threshold)? 1: 0;
    return e;
}

//--------------------------------------------------------------------------
float4 edgeDetectPS( uniform bool combine, PostProc_VSOut IN ): SV_TARGET
{
    // We need eight samples (the centre has zero weight in both kernels).
    float3 offset = {1, -1 , 0};

    float g00 = tDepth.Sample( samNearest, IN.tex.xy + offset.yy * g_InvResolution).x;
    float g01 = tDepth.Sample( samNearest, IN.tex.xy + offset.zy * g_InvResolution).x;
    float g02 = tDepth.Sample( samNearest, IN.tex.xy + offset.xy * g_InvResolution).x;
    float g10 = tDepth.Sample( samNearest, IN.tex.xy + offset.yz * g_InvResolution).x;
    float g11 = tDepth.Sample( samNearest, IN.tex.xy + offset.zz * g_InvResolution).x;
    float g12 = tDepth.Sample( samNearest, IN.tex.xy + offset.xz * g_InvResolution).x;
    float g20 = tDepth.Sample( samNearest, IN.tex.xy + offset.yx * g_InvResolution).x;
    float g21 = tDepth.Sample( samNearest, IN.tex.xy + offset.zx * g_InvResolution).x;
    float g22 = tDepth.Sample( samNearest, IN.tex.xy + offset.xx * g_InvResolution).x;

    // Sobel in horizontal dir.
    float4 sx = 0;
    sx -= g00;
    sx -= g01 * 2;
    sx -= g02;
    sx += g20;
    sx += g21 * 2;
    sx += g22;
    
    // Sobel in vertical dir - weights are just rotated 90 degrees.
    float4 sy = 0;
    sy -= g00;
    sy += g02;
    sy -= g10 * 2;
    sy += g12 * 2;
    sy -= g20;
    sy += g22;
    
    //return g11;

    // In theory, dist should use a sqrt.  This is a common approx.
    //float greySx = 0.333 * (sx.r + sx.g + sx.b);
    //float greySy = 0.333 * (sy.r + sy.g + sy.b);
    //float eR = edgeDetectScalar(greySx, greySy);
    if( !edgeDetectScalar(sx, sy, g_EdgeThreshold)){
        discard;    
    }
    
    return 1;
    
}

//-------------------------------------------------------------------------
technique10 BlurEdgeDetection
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, edgeDetectPS(false) ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
    }
}
