//----------------------------------------------------------------------------------
// File:   RayMarchingAOEngine.cpp
// Author: Miguel Sainz &  Louis Bavoil
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

#include "RayMarchingAOEngine.h"

#define MAX_PATH_STR 512

enum {
    IDC_TR_STATIC = 4,
    IDC_TRACE_RADIUS,
    IDC_ND_STATIC,
    IDC_NUM_DIR,
    IDC_NR_STATIC,
    IDC_NUM_RAY,
    IDC_NS_STATIC,
    IDC_NUM_STEPS,
    IDC_PRESET_STATIC,
    IDC_PRESET_COMBO,
    IDC_LA_STATIC,
    IDC_LINEAR_ATT,
    IDC_CONTRAST,
    IDC_CONTRAST_STATIC,
    IDC_BG_STATIC,
};

enum {
    PRESET_QUALITY,
    PRESET_FINE,
    PRESET_STANDARD,
    PRESET_FAST
};

struct sPreset
{
    float Radius;
    int   Dirs;
    int   Rays;
    int   Steps;
    float LinAtt;
    float Contrast;
};

const sPreset PRESETS[4] =
{
    {0.35f, 16, 8, 16, 1.0f, 2.0f},
    {0.35f,  8, 6,  8, 1.0f, 2.0f},
    {0.2f,   7, 2,  6, 1.0f, 2.0f},
    {0.15f,  6, 2,  3, 1.0f, 2.0f}
};

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
RayMarchingAOEngine::RayMarchingAOEngine()
{
    m_BBWidth    = 0;
    m_BBHeight   = 0;

    m_Effect = 0;

    m_SSAOBuffer	= NULL; 
    m_D3DDevice		= NULL;
    m_pRndTexture	= NULL;
    m_pRVRndTex     = NULL;

    m_TraceRadius	= PRESETS[PRESET_FINE].Radius;
    m_NumDirs		= PRESETS[PRESET_FINE].Dirs;
    m_NumRays		= PRESETS[PRESET_FINE].Rays;
    m_NumSteps		= PRESETS[PRESET_FINE].Steps;

    m_LinAtt        = PRESETS[PRESET_FINE].LinAtt;
    m_Contrast      = PRESETS[PRESET_FINE].Contrast;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT RayMarchingAOEngine::OnResizedSwapChain( int width, int height, float fovy )                                     
{
    HRESULT hr = S_OK;

     // Recalculate buffer sizes
    m_BBWidth    = width;
    m_BBHeight   = height;

    // Create the RT for HDR with MSAA
    D3D10_TEXTURE2D_DESC pDesc;
    pDesc.Width              = m_BBWidth;
    pDesc.Height             = m_BBHeight;
    pDesc.MipLevels          = 1;
    pDesc.ArraySize          = 1;
    pDesc.Format             = DXGI_FORMAT_R8_UNORM;
    pDesc.SampleDesc.Count   = 1;
    pDesc.SampleDesc.Quality = 0;
    pDesc.Usage              = D3D10_USAGE_DEFAULT;
    pDesc.BindFlags          = D3D10_BIND_SHADER_RESOURCE | D3D10_BIND_RENDER_TARGET;
    pDesc.CPUAccessFlags     = 0;
    pDesc.MiscFlags          = 0;

    SAFE_DELETE(m_SSAOBuffer);
    m_SSAOBuffer = new SimpleRT( m_D3DDevice , &pDesc);

    pDesc.Format             = DXGI_FORMAT_R32G32B32A32_FLOAT;

    m_FocalLen[0]      = 1.0f / tanf(fovy * 0.5f) *  (float)m_BBHeight / (float)m_BBWidth;
    m_FocalLen[1]      = 1.0f / tanf(fovy * 0.5f);
    m_InvResolution[0] = 1.0f / m_BBWidth;
    m_InvResolution[1] = 1.0f / m_BBHeight;

    m_pFocalLen->SetFloatVector( m_FocalLen );
    m_pInvResolution->SetFloatVector( m_InvResolution );

    if( FAILED( hr ) ){ 
        MessageBox(NULL, L"Initialization of RayMarchingAOEngine failed!!!", L"ERROR", MB_OK|MB_SETFOREGROUND|MB_TOPMOST);		
        return FALSE;
    }
    return hr;
}
//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void RayMarchingAOEngine::OnReleasingSwapChain( )
{
    m_BBWidth  = 0;
    m_BBHeight = 0;
    
    SAFE_DELETE(m_SSAOBuffer);
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT RayMarchingAOEngine::OnDestroyDevice( )
{
    SAFE_RELEASE(m_Effect);
    SAFE_RELEASE(m_pRVRndTex);
    SAFE_RELEASE(m_pRndTexture);
    return S_OK;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT RayMarchingAOEngine::OnFrameRender( ID3D10ShaderResourceView *Depth, ID3D10ShaderResourceView *Normal )
{
    HRESULT hr = S_OK;
    
    m_D3DDevice->OMSetRenderTargets( 1, &m_SSAOBuffer->pRTV, NULL );

    m_DepTex->SetResource( Depth );
    m_NorTex->SetResource( Normal );
    m_RanTex->SetResource( m_pRVRndTex );

    m_Technique_RAYMARCHING_AO->GetPassByIndex(0)->Apply( 0 );
    
    m_D3DDevice->Draw( 3, 0 );

    m_DepTex->SetResource( NULL );
    m_NorTex->SetResource( NULL );
    m_RanTex->SetResource( NULL );

    m_Technique_RAYMARCHING_AO->GetPassByIndex(0)->Apply( 0 );

    return hr;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void RayMarchingAOEngine::CreateMenu ( CDXUTDialog *myMenu )
{
    int iY = 0;
    WCHAR sz[100];
    CDXUTComboBox *pComboBox;
    WCHAR* szPresets[] =
    {
        L"Quality",
        L"Fine",
        L"Performance",
        L"Fast",
    };

    StringCchPrintf( sz, 100, L"Preset:" );
    myMenu->AddStatic( IDC_PRESET_STATIC, sz, 35, iY += 24 + 20, 125, 22 );
    myMenu->AddComboBox( IDC_PRESET_COMBO,    35, iY += 24, 125, 22, 'T', false, &pComboBox);

    for( char i = 0; i < 4; i++ ){
        pComboBox->AddItem( szPresets[i], (void*)i );
    }
    pComboBox->SetSelectedByIndex(PRESET_FINE);

    int dy = 22;
    StringCchPrintf( sz, 100, L"Radius: %0.2f", m_TraceRadius ); 
    myMenu->AddStatic( IDC_TR_STATIC, sz, 35, iY += dy + 15, 125, 22 );
    myMenu->AddSlider( IDC_TRACE_RADIUS, 50, iY += dy, 100, 22, 0, 100, (int)(m_TraceRadius*200.0f) );

    StringCchPrintf( sz, 100, L"Number of directions: %d", m_NumDirs ); 
    myMenu->AddStatic( IDC_ND_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_NUM_DIR, 50, iY += dy, 100, 22, 0, 100, 4*m_NumDirs );

    StringCchPrintf( sz, 100, L"Number of rays/dir: %d", m_NumRays ); 
    myMenu->AddStatic( IDC_NR_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_NUM_RAY, 50, iY += dy, 100, 22, 0, 100, 4*m_NumRays );

    StringCchPrintf( sz, 100, L"Number of steps: %d", m_NumSteps ); 
    myMenu->AddStatic( IDC_NS_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_NUM_STEPS, 50, iY += dy, 100, 22, 0, 128, 4*m_NumSteps );

    StringCchPrintf( sz, 100, L"Linear attenuation: %0.2f", m_LinAtt ); 
    myMenu->AddStatic( IDC_LA_STATIC, sz, 35, iY += dy, 125, 22 );
    myMenu->AddSlider( IDC_LINEAR_ATT, 50, iY += dy, 100, 22, 0, 100, (int)(100.0f*m_LinAtt) );

    StringCchPrintf( sz, 100, L"Contrast: %0.2f", m_Contrast );
    myMenu->AddStatic( IDC_CONTRAST_STATIC, sz, 35, iY += 24, 125, 22 );
    myMenu->AddSlider( IDC_CONTRAST, 50, iY += 24, 100, 22, 0, 100, (int)(50.0f*m_Contrast) );
}

void RayMarchingAOEngine::OnGUIEvent( CDXUTDialog *myMenu, UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {
        case IDC_PRESET_COMBO:
        {
            WCHAR sz[100];
            CDXUTComboBox* pComboBox = (CDXUTComboBox*) pControl; 
            int i = PtrToInt(pComboBox->GetSelectedData());
            m_TraceRadius	= PRESETS[i].Radius;	m_pTraceRadius->SetFloat( m_TraceRadius );
            m_NumDirs		= PRESETS[i].Dirs;		m_pNumDirs->SetInt( m_NumDirs );
            m_NumSteps		= PRESETS[i].Steps; 	m_pNumSteps->SetInt( m_NumSteps );

            StringCchPrintf( sz, 100, L"Trace radius: %0.2f", m_TraceRadius );
            myMenu->GetStatic( IDC_TR_STATIC )->SetText( sz );
            myMenu->GetSlider( IDC_TRACE_RADIUS )->SetValue( (int)(m_TraceRadius * 200.0f) );

            StringCchPrintf( sz, 100, L"Number of directions: %d", m_NumDirs ); 
            myMenu->GetStatic( IDC_ND_STATIC )->SetText( sz );
            myMenu->GetSlider( IDC_NUM_DIR )->SetValue( m_NumDirs * 4 );

            StringCchPrintf( sz, 100, L"Number of steps: %d", m_NumSteps ); 
            myMenu->GetStatic( IDC_NS_STATIC )->SetText( sz );
            myMenu->GetSlider( IDC_NUM_STEPS )->SetValue( m_NumSteps * 4 );

            UpdateDirs();
            break;
        }    
        case IDC_TRACE_RADIUS: 
        {
            m_TraceRadius = (float) myMenu->GetSlider( IDC_TRACE_RADIUS )->GetValue() * 0.005f;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Trace radius: %0.2f", m_TraceRadius ); 
            myMenu->GetStatic( IDC_TR_STATIC )->SetText( sz );

            m_pTraceRadius->SetFloat( m_TraceRadius );
            break;
        }
        case IDC_NUM_DIR: 
        {
            m_NumDirs = myMenu->GetSlider( IDC_NUM_DIR )->GetValue() / 4;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Number of directions: %d", m_NumDirs ); 
            myMenu->GetStatic( IDC_ND_STATIC )->SetText( sz );

            m_pNumDirs->SetInt( m_NumDirs );
            UpdateDirs();
            break;
        }
        case IDC_NUM_RAY: 
        {
            m_NumRays = myMenu->GetSlider( IDC_NUM_RAY )->GetValue() / 4;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Number of rays/dir: %d", m_NumRays ); 
            myMenu->GetStatic( IDC_NR_STATIC )->SetText( sz );

            m_pNumRays->SetInt( m_NumRays );
            break;
        }
        case IDC_NUM_STEPS: 
        {
            m_NumSteps = myMenu->GetSlider( IDC_NUM_STEPS )->GetValue() / 4;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Number of steps: %d", m_NumSteps ); 
            myMenu->GetStatic( IDC_NS_STATIC )->SetText( sz );

            m_pNumSteps->SetInt( m_NumSteps );
            break;
        }
        case IDC_CONTRAST: 
        {
            m_Contrast = (float)myMenu->GetSlider( IDC_CONTRAST )->GetValue() / 50.0f;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Contrast: %0.2f", m_Contrast );  
            myMenu->GetStatic( IDC_CONTRAST_STATIC )->SetText( sz );

            m_pContrast->SetFloat( m_Contrast );
            break;
        }
        case IDC_LINEAR_ATT: 
        {
            m_LinAtt = (float)myMenu->GetSlider( IDC_LINEAR_ATT )->GetValue() / 100.0f;

            WCHAR sz[100];
            StringCchPrintf( sz, 100, L"Linear attenuation: %0.2f", m_LinAtt ); 
            myMenu->GetStatic( IDC_LA_STATIC )->SetText( sz );

            m_pLinAtt->SetFloat( m_LinAtt );
            break;
        }
    }
}

void RayMarchingAOEngine::UpdateDirs()
{
    float dirs[64];
    if(m_NumDirs > 32)
        m_NumDirs = 32;
    float inc = 2.0f*D3DX_PI/(float)m_NumDirs;
    for(int i=0; i<2*m_NumDirs; i+=2)
    {
        float angle = inc*(float)(i/2);
        dirs[i] = cos(angle);
        dirs[i+1] = sin(angle);
    }

    m_pNumDirs->SetInt(m_NumDirs);
    m_pDirsArray->SetFloatVectorArray(dirs,0,m_NumDirs);

    SAFE_RELEASE(m_pRndTexture);
    SAFE_RELEASE(m_pRVRndTex);

#define SCALE ((1<<15))

    signed short f[64*64*4];
    for(int i=0; i<64*64*4; i+=4)
    {
        float rnd = (float)rand();
        float angle = 2.0f*D3DX_PI * rnd / (float)RAND_MAX / (float)m_NumDirs;
        f[i  ] = (signed short)(SCALE*cos(angle));
        f[i+1] = (signed short)(SCALE*sin(angle));
        f[i+2] = (signed short)(SCALE*(float)(rnd / RAND_MAX));
        f[i+3] = 0;
    }

    D3D10_TEXTURE2D_DESC tex_desc;
    tex_desc.Width = 64;
    tex_desc.Height = 64;
    tex_desc.MipLevels = tex_desc.ArraySize = 1;
    tex_desc.Format = DXGI_FORMAT_R16G16B16A16_SNORM;
    tex_desc.SampleDesc.Count = 1;
    tex_desc.SampleDesc.Quality = 0;
    tex_desc.Usage = D3D10_USAGE_IMMUTABLE;
    tex_desc.BindFlags = D3D10_BIND_SHADER_RESOURCE;
    tex_desc.CPUAccessFlags = 0;
    tex_desc.MiscFlags      = 0;

    D3D10_SUBRESOURCE_DATA sr_desc;
    sr_desc.pSysMem = f;
    sr_desc.SysMemPitch = 64*4*sizeof(signed short);
    sr_desc.SysMemSlicePitch = 0;

    D3D10_SHADER_RESOURCE_VIEW_DESC srv_desc;
    srv_desc.Format = tex_desc.Format;
    srv_desc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
    srv_desc.Texture2D.MostDetailedMip = 0;
    srv_desc.Texture2D.MipLevels = 1;

    m_D3DDevice->CreateTexture2D(&tex_desc, &sr_desc, &m_pRndTexture);

    m_D3DDevice->CreateShaderResourceView(m_pRndTexture, &srv_desc, &m_pRVRndTex);
}

HRESULT RayMarchingAOEngine::OnCreateDevice( ID3D10Device* pd3dDevice )
{
    HRESULT hr = S_OK;

    m_D3DDevice = pd3dDevice;

    RecompileShader();
    UpdateDirs();

    srand(57);

    return hr;
}

RayMarchingAOEngine::~RayMarchingAOEngine()
{
    SAFE_RELEASE(m_pRndTexture);
    SAFE_RELEASE(m_pRVRndTex);
}

void RayMarchingAOEngine::RecompileShader()
{
    ID3D10Effect *tempEffect = NULL;

    // Create the effect
    WCHAR str[MAX_PATH_STR+1];
    
    DXUTFindDXSDKMediaFileCch( str, MAX_PATH_STR, DEFAULT_SHADERPATH L"\\ScreenSpaceAO\\RayMarchingAOEngine.fx" );
    D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, m_D3DDevice, NULL, NULL, &tempEffect, NULL, NULL);
    if( tempEffect && tempEffect->IsValid() ){
        SAFE_RELEASE(m_Effect);
        m_Effect = tempEffect;
    }else{
        MessageBox(NULL, L"Effect Compilation failed. Review your code a*#$%le", L"ERROR", MB_OK|MB_SETFOREGROUND|MB_TOPMOST);	
        return;
    }
    
    // Obtain the technique
    m_Technique_RAYMARCHING_AO  = m_Effect->GetTechniqueByName( "RAYMARCHING_AO_Pass" );
    
    // Obtain pointers to the shader variales
    m_DepTex      = m_Effect->GetVariableByName( "tLinDepth" )->AsShaderResource();
    m_NorTex      = m_Effect->GetVariableByName( "tNormal"   )->AsShaderResource();
    m_RanTex      = m_Effect->GetVariableByName( "tRandom"   )->AsShaderResource();

    m_pNumSteps = m_Effect->GetVariableByName     ( "g_NumSteps" )->AsScalar();
    m_pTraceRadius = m_Effect->GetVariableByName  ( "g_R" )->AsScalar();
    m_pNumDirs = m_Effect->GetVariableByName      ( "g_NumDirs" )->AsScalar();
    m_pNumRays = m_Effect->GetVariableByName      ( "g_NumRays" )->AsScalar();
    m_pLinAtt = m_Effect->GetVariableByName       ( "g_LinAtt" )->AsScalar();
    m_pContrast = m_Effect->GetVariableByName     ( "g_Contrast" )->AsScalar();
    m_pFocalLen = m_Effect->GetVariableByName     ( "g_FocalLen" )->AsVector();
    m_pDirsArray = m_Effect->GetVariableByName    ( "g_Dirs" )->AsVector();
    m_pInvResolution = m_Effect->GetVariableByName( "g_InvResolution" )->AsVector();

    // Set Defaults
    UpdateDirs();
        
    m_pTraceRadius->SetFloat(m_TraceRadius);
    m_pNumSteps->SetInt(m_NumSteps);
    m_pNumRays->SetInt(m_NumRays);
    m_pLinAtt->SetFloat(m_LinAtt);
    m_pContrast->SetFloat(m_Contrast);

    m_pFocalLen->SetFloatVector( m_FocalLen );
    m_pInvResolution->SetFloatVector( m_InvResolution );
}
