//----------------------------------------------------------------------------------
// File:    RayMarchingAOEngine.fx
// Author:  Miguel Sainz, Louis Bavoil, Rouslan Dimitrov
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//----------------------------------------------------------------------------------

#define COSINE_WEIGHTED

Texture2D<float3> tRandom;
Texture2D<float>  tLinDepth;
Texture2D<float4> tNormal;

SamplerState samNearest
{
    Filter   = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

//----------------------------------------------------------------------------------
DepthStencilState DisableDepth
{
    DepthEnable    = FALSE;
    DepthWriteMask = ZERO;
};

BlendState DisableBlend
{
    BlendEnable[0] = false;
};

//----------------------------------------------------------------------------------
cbuffer cb0
{
    float2 g_Dirs[32];
}

cbuffer cb1
{
    float  g_NumSteps;
    float  g_R;
    int    g_NumDirs;
    float  g_NumRays;

    float  g_LinAtt;
    float  g_Contrast;
    float2 g_FocalLen;
    float2 g_InvResolution;
}

//----------------------------------------------------------------------------------
struct PostProc_VSOut
{
    float4 pos   : SV_Position;
    float2 tex   : TEXCOORD0;
    float2 texUV : TEXCOORD1;
    float2 texXY : TEXCOORD2;
};

//Vertex shader that generates a full screen quad with texcoords g_Rom vertIDs
//To use draw 3 vertices with primitive type triangle strip
PostProc_VSOut FullScreenQuadVS( uint id : SV_VertexID )
{
    PostProc_VSOut output = (PostProc_VSOut)0.0f;
    float2 tex = float2( (id << 1) & 2, id & 2 );
    output.tex = tex * float2( 2.0f, -2.0f ) + float2( -1.0f, 1.0f);
    output.pos = float4( output.tex, 0.0f, 1.0f );
    output.tex /= g_FocalLen;
    
    // Bottom left pixel is (0,0) and bottom right is (1,1)
    output.texUV = float2( (id << 1) & 2, id & 2 );
    output.texXY = (float2(2.0, -2.0) * (output.texUV - float2(0.5, 0.5)));
    
    return output;
}

//----------------------------------------------------------------------------------
float3 fetch_eye_pos(float3 pos)
{
    float2 r =  pos.xy / pos.z;
    float2 tx_n = float2(0.5, -0.5)*(g_FocalLen * r + float2(1.0, -1.0));
    
    float z = tLinDepth.SampleLevel(samNearest, float3(tx_n.x, tx_n.y, 0), 0).x;
    return float3(z*r.x, z*r.y, z);
}

//----------------------------------------------------------------------------------
float4 RAYMARCHING_AO_PS ( PostProc_VSOut IN ) : SV_TARGET
{
    float z = tLinDepth.SampleLevel(samNearest, float3(IN.texUV, 0), 0).x;
    float3 P = float3(z*IN.tex.xy, z);

    // Account for far plane
    if(P.z > 100)
        return 1;

    // Get the basis per pixel
    float3 N = normalize(tNormal.SampleLevel(samNearest, float3(IN.texUV, 0), 0).xyz);
    float3 Tan   = float3(1,0,0);
    float3 BiTan = normalize(cross(N, Tan));
    Tan          = cross(BiTan, N);

    const float step_size = g_R / g_NumSteps;
    float3 rand = tRandom.Load(int3((int)IN.pos.x&63, (int)IN.pos.y&63, 0)).xyz;
    
    float3 dir_t;
    float color = 0.0;
    for (int d = 0; d < g_NumDirs; d++) {
        float3 dir = float3(g_Dirs[d].x*rand.x - g_Dirs[d].y*rand.y, 
                            g_Dirs[d].x*rand.y + g_Dirs[d].y*rand.x, 
                            0);

        dir = dir.x*Tan + dir.y*BiTan;

        float n_weight = 1.0/(g_NumRays*g_NumDirs);
        for (float n = 1; n <= g_NumRays; n++) {
            float frac = n/(g_NumRays + 1.5e-2);
#ifdef COSINE_WEIGHTED
            float3 ndir = dir*sqrt(frac) + N*sqrt(1-frac);		
#else
            float3 ndir = dir*frac + N*sqrt(1.0-frac*frac);
#endif

            for (float i = 1.0; i <= g_NumSteps; i++) {
                float3 cur_ray = (i+rand.z)*step_size*ndir;
                float3 cur_pos = cur_ray + P;
                float3 tex_pos = fetch_eye_pos(cur_pos);
        
                if (tex_pos.z - cur_pos.z < 0.0) {
                    float l = length(P-tex_pos);
                    if (l < g_R) {
                        color -= n_weight * (g_R-g_LinAtt*l)/g_R;					
                        break;
                    }
                }
            }
        }
    }

    return 1.0 + color * g_Contrast;
}

//----------------------------------------------------------------------------------
technique10 RAYMARCHING_AO_Pass
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, RAYMARCHING_AO_PS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepth, 0 );

    }
}

