//----------------------------------------------------------------------------------
// File:    ResolveEngine.cpp
// Authors: Louis Bavoil & Miguel Sainz
// Email:   sdkfeedback@nvidia.com
// 
// Copyright (c) 2008 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

#include "ResolveEngine.h"

#define MAX_PATH_STR 512

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
ResolveEngine::ResolveEngine()
{
    m_inWidth       = 0;
    m_inHeight      = 0;

    m_outWidth      = 0;
    m_outHeight     = 0;

    m_Effect        = NULL;
    m_D3DDevice		= NULL;

    m_ColorTexture  = NULL;
    m_NormalTexture = NULL;
    m_DepthTexture  = NULL;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT ResolveEngine::OnResizedSwapChain( int inputWidth,  int inputHeight, int nsamples,   
                                           int outputWidth, int outputHeight, 
                                           float zFar, float zNear )
{
    HRESULT hr = S_OK;

    m_inWidth    = inputWidth;
    m_inHeight   = inputHeight;
    m_outWidth   = outputWidth;
    m_outHeight  = outputHeight;
    m_BBNSamples = nsamples;
    m_ZFar       = zFar;
    m_ZNear      = zNear;

     // Do we require upsampling?
    if( m_inWidth != m_outWidth || m_inHeight != m_outHeight )  m_NeedUpsampling = true;
    else                                                        m_NeedUpsampling = false;

    RecompileShader();

    m_pZFar->SetFloat( m_ZFar );
    m_pZNear->SetFloat( m_ZNear );

    // Create resolved (non-MSAA) RTs
    D3D10_TEXTURE2D_DESC texDesc;
    texDesc.Width              = m_inWidth;
    texDesc.Height             = m_inHeight;
    texDesc.MipLevels          = 1;
    texDesc.ArraySize          = 1;
    texDesc.SampleDesc.Count   = 1;
    texDesc.SampleDesc.Quality = 0;
    texDesc.Usage              = D3D10_USAGE_DEFAULT;
    texDesc.BindFlags          = D3D10_BIND_SHADER_RESOURCE | D3D10_BIND_RENDER_TARGET;
    texDesc.CPUAccessFlags     = 0;
    texDesc.MiscFlags          = 0;

    texDesc.Format             = DXGI_FORMAT_R8G8B8A8_UNORM;
    SAFE_DELETE(m_ColorTexture);
    m_ColorTexture = new SimpleRT( m_D3DDevice, &texDesc );

    texDesc.Width              = m_outWidth;
    texDesc.Height             = m_outHeight;

    texDesc.Format             = DXGI_FORMAT_R8G8B8A8_SNORM;
    SAFE_DELETE(m_NormalTexture);
    m_NormalTexture = new SimpleRT( m_D3DDevice, &texDesc );

    texDesc.Format             = DXGI_FORMAT_R32_FLOAT;
    SAFE_DELETE(m_DepthTexture);
    m_DepthTexture = new SimpleRT( m_D3DDevice, &texDesc );

    SAFE_DELETE(m_SrcDepthTexture);
    if( m_NeedUpsampling )
    {
        texDesc.Width  = m_inWidth;
        texDesc.Height = m_inHeight;
        texDesc.Format = DXGI_FORMAT_R32_FLOAT;
        m_SrcDepthTexture = new SimpleRT( m_D3DDevice, &texDesc );
    }
    else
    {
        m_SrcDepthTexture = NULL;
    }

    m_viewportInput.TopLeftX = 0;
    m_viewportInput.TopLeftY = 0;
    m_viewportInput.MinDepth = 0;
    m_viewportInput.MaxDepth = 1;
    m_viewportInput.Width    = m_inWidth;
    m_viewportInput.Height   = m_inHeight;

    m_viewportOutput.TopLeftX = 0;
    m_viewportOutput.TopLeftY = 0;
    m_viewportOutput.MinDepth = 0;
    m_viewportOutput.MaxDepth = 1;
    m_viewportOutput.Width    = m_outWidth;
    m_viewportOutput.Height   = m_outHeight;

    if( FAILED( hr ) )
    { 
        MessageBox(NULL, L"Initialization of ResolveEngine failed!!!", L"ERROR", MB_OK|MB_SETFOREGROUND|MB_TOPMOST);		
        return FALSE;
    }
    return hr;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void ResolveEngine::OnReleasingSwapChain( )
{
    m_inWidth  = 0;
    m_inHeight = 0;

    m_outWidth  = 0;
    m_outHeight = 0;
    
    SAFE_DELETE(m_ColorTexture);
    SAFE_DELETE(m_NormalTexture);
    SAFE_DELETE(m_DepthTexture);
    SAFE_DELETE(m_SrcDepthTexture);
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT ResolveEngine::OnDestroyDevice( )
{
    SAFE_RELEASE(m_Effect);
    return S_OK;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
ID3D10ShaderResourceView* ResolveEngine::LinearizeDepth(ID3D10ShaderResourceView* DepthBufferSRV)
{
    SimpleRT *destRT = m_NeedUpsampling ? m_SrcDepthTexture : m_DepthTexture;

    m_D3DDevice->RSSetViewports(1, &m_viewportInput );
    m_D3DDevice->OMSetRenderTargets( 1, &destRT->pRTV, NULL );
   
    m_DepthTex->SetResource( DepthBufferSRV );
    m_TechniqueLinearizeD->GetPassByIndex( 0 )->Apply( 0 );

    // Draw fullscreen triangle
    m_D3DDevice->Draw( 3, 0 );

    m_DepthTex->SetResource( NULL );
    m_TechniqueLinearizeD->GetPassByIndex( 0 )->Apply( 0 );

    return destRT->pSRV;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
void ResolveEngine::ResolveDepthsAndNormals(ID3D10ShaderResourceView* FullResLinearDepth,
                                            ID3D10ShaderResourceView* MRTNormals)
{
    if (MRTNormals)
    {
        ID3D10RenderTargetView* MRTs[2] = { m_NormalTexture->pRTV, m_DepthTexture->pRTV };
        m_D3DDevice->OMSetRenderTargets( 2, MRTs, NULL );

        m_NormalTex->SetResource(MRTNormals);
        m_DepthTex->SetResource(FullResLinearDepth);
        m_TechniqueND->GetPassByIndex( 0 )->Apply( 0 );
    }
    else
    {
        m_D3DDevice->OMSetRenderTargets( 1, &m_DepthTexture->pRTV, NULL );
        m_DepthTex->SetResource(FullResLinearDepth);
        m_TechniqueD->GetPassByIndex( 0 )->Apply( 0 );    
    }

    // Draw fullscreen triangle
    m_D3DDevice->Draw( 3, 0 );

    m_NormalTex->SetResource(NULL);
    m_DepthTex->SetResource(NULL);
    m_TechniqueND->GetPassByIndex( 0 )->Apply( 0 );
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT	ResolveEngine::ResolveColors( SimpleRT* MRTColor,
                                      ID3D10ShaderResourceView** ResolvedColorSRV )
{
    if (m_BBNSamples == 1)
    {
        *ResolvedColorSRV = MRTColor->pSRV;
    }
    else
    {
        m_D3DDevice->ResolveSubresource((ID3D10Resource*)m_ColorTexture->pTexture, 0, (ID3D10Resource*)MRTColor->pTexture, 0, DXGI_FORMAT_R8G8B8A8_UNORM);
        *ResolvedColorSRV = m_ColorTexture->pSRV;
    }
    return S_OK;
}

//--------------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
HRESULT	ResolveEngine::ResolveNormalsAndDepths( ID3D10ShaderResourceView*  MRTNormals,
                                                ID3D10ShaderResourceView*  MRTDepths,
                                                ID3D10ShaderResourceView*  DepthBufferSRV,
                                                ID3D10ShaderResourceView** ResolvedNormalSRV,
                                                ID3D10ShaderResourceView** ResolvedDepthSRV,
                                                ID3D10ShaderResourceView** FullResLinearDepth)
{
    // If we have a depth buffer, we will need to linearize it in full rez, and resolve at the same time
    // regardless of whether we need to downsample or not, so let's take care of that here
    if (!MRTDepths)
    {
        if (m_BBNSamples == 1)
        {
            *FullResLinearDepth = LinearizeDepth(DepthBufferSRV);
        }
        else
        {
            // DX10.0 does not allow fetching from multisample depth buffers
            *ResolvedNormalSRV  = NULL;
            *ResolvedDepthSRV   = NULL;
            *FullResLinearDepth = NULL;
            return S_FALSE;
        }
    }
    else
    {
        *FullResLinearDepth = MRTDepths;
    }

    // Non-MSAA case
    if (m_BBNSamples == 1)
    {
        if (!m_NeedUpsampling)
        {
            *ResolvedNormalSRV = MRTNormals;
            *ResolvedDepthSRV = *FullResLinearDepth;
        }
        else
        {
            m_D3DDevice->RSSetViewports(1, &m_viewportOutput );
            ResolveDepthsAndNormals(*FullResLinearDepth, MRTNormals);
            *ResolvedNormalSRV = m_NormalTexture->pSRV;
            *ResolvedDepthSRV  = m_DepthTexture->pSRV;
        }
        return S_OK;
    }

    // MSAA case
    if (m_BBNSamples > 1)
    {
        if(!m_NeedUpsampling){
            m_D3DDevice->RSSetViewports(1, &m_viewportOutput );
        }else{
            m_D3DDevice->RSSetViewports(1, &m_viewportInput );
        }
        ResolveDepthsAndNormals(*FullResLinearDepth, MRTNormals);
        *ResolvedNormalSRV = MRTNormals ? m_NormalTexture->pSRV : NULL;
        *ResolvedDepthSRV  = m_DepthTexture->pSRV;            
    }

    return S_OK;

}

HRESULT ResolveEngine::OnCreateDevice( ID3D10Device* pd3dDevice )
{
    HRESULT hr = S_OK;

    m_D3DDevice = pd3dDevice;
    
    return hr;
}

ResolveEngine::~ResolveEngine()
{
}

void ResolveEngine::RecompileShader()
{
    ID3D10Effect *tempEffect = NULL;

    D3D10_SHADER_MACRO Shader_Macros[3];
    memset(Shader_Macros, 0, sizeof(Shader_Macros));

    CHAR defineStr[2];
    StringCchPrintfA(defineStr, 2, "%d", m_BBNSamples);
    Shader_Macros[0].Name = "NUM_MSAA_SAMPLES";
    Shader_Macros[0].Definition = defineStr;
    
    Shader_Macros[1].Name = "DOWNSAMPLED";
    if (m_NeedUpsampling) Shader_Macros[1].Definition = "1";
    else                  Shader_Macros[1].Definition = "0";

    // Create the effect
    WCHAR str[MAX_PATH_STR+1];
    DXUTFindDXSDKMediaFileCch( str, MAX_PATH_STR, DEFAULT_SHADERPATH L"\\ScreenSpaceAO\\ResolveEngine.fx" );
    D3DX10CreateEffectFromFile( str, Shader_Macros, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, m_D3DDevice, NULL, NULL, &tempEffect, NULL, NULL);
    if( tempEffect && tempEffect->IsValid() )
    {
        SAFE_RELEASE(m_Effect);
        m_Effect = tempEffect;
    }
    else
    {
        MessageBox(NULL, L"Effect Compilation failed. Review your code a*#$%le", L"ERROR", MB_OK|MB_SETFOREGROUND|MB_TOPMOST);	
        return;
    }
    
    // Obtain the technique
    m_TechniqueND            = m_Effect->GetTechniqueByName( "ResolveND" );
    m_TechniqueD             = m_Effect->GetTechniqueByName( "ResolveD" );
    m_TechniqueLinearizeD    = m_Effect->GetTechniqueByName( "LinearizeDepth" );

    // Obtain pointers to the shader resources
    m_NormalTex      = m_Effect->GetVariableByName( "tNormal"  )->AsShaderResource();
    m_DepthTex       = m_Effect->GetVariableByName( "tDepth"   )->AsShaderResource();
    m_pZNear         = m_Effect->GetVariableByName( "g_ZNear" )->AsScalar();
    m_pZFar          = m_Effect->GetVariableByName( "g_ZFar"  )->AsScalar();

    m_pZNear->SetFloat( m_ZNear );
    m_pZFar->SetFloat( m_ZFar );
}
