//--------------------------------------------------------------------------------------
// File:    ResolveEngine.fx
// Authors: Louis Bavoil & Miguel Sainz
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#ifndef NUM_MSAA_SAMPLES
#define NUM_MSAA_SAMPLES 1
#endif

#ifndef DOWNSAMPLED
#define DOWNSAMPLED 0
#endif

#if NUM_MSAA_SAMPLES == 1
Texture2D<float3> tNormal;
Texture2D<float>  tDepth;
#else
Texture2DMS<float3, NUM_MSAA_SAMPLES> tNormal;
Texture2DMS<float, NUM_MSAA_SAMPLES>  tDepth;
#endif

float g_ZNear;
float g_ZFar;

SamplerState samNearest
{
    Filter   = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

//--------------------------------------------------------------------------------------
DepthStencilState DisableDepth
{
    DepthEnable    = FALSE;
    DepthWriteMask = ZERO;
};

BlendState DisableBlend
{
    BlendEnable[0] = false;
};

struct PS_OUTPUT
{
    float3 Normal : SV_Target0;
    float  Depth  : SV_Target1;
};

//--------------------------------------------------------------------------------------
struct PostProc_VSOut
{
    float4 pos : SV_Position;
    float2 tex : TEXCOORD0;
};

//Vertex shader that generates a full screen quad with texcoords g_Rom vertIDs
//To use draw 3 vertices with primitive type triangle strip
PostProc_VSOut FullScreenQuadVS( uint id : SV_VertexID )
{
    PostProc_VSOut output = (PostProc_VSOut)0.0f;

    output.tex = float2( (id << 1) & 2, id & 2 );
    output.pos = float4( output.tex * float2( 2.0f, -2.0f ) + float2( -1.0f, 1.0f) , 0.0f, 1.0f );
    
    return output;
}

//-------------------------------------------------------------------------
PS_OUTPUT ResolveND_PS( PostProc_VSOut IN)
{
    PS_OUTPUT output;
#if NUM_MSAA_SAMPLES == 1
    output.Normal = tNormal.SampleLevel( samNearest, IN.tex, 0 );
    output.Depth  = tDepth.SampleLevel ( samNearest, IN.tex, 0 );
#else

#if DOWNSAMPLED == 1
    output.Normal = tNormal.Load( int2(IN.pos.xy)*2, 0 );
    output.Depth  = tDepth.Load ( int2(IN.pos.xy)*2, 0 );
#else
    output.Normal = tNormal.Load( int2(IN.pos.xy), 0 );
    output.Depth  = tDepth.Load ( int2(IN.pos.xy), 0 );
#endif

#endif
    return output;
}

//-------------------------------------------------------------------------
float ResolveD_PS( PostProc_VSOut IN) : SV_TARGET
{
#if NUM_MSAA_SAMPLES == 1
    return tDepth.SampleLevel( samNearest, IN.tex, 0 );
#else

#if DOWNSAMPLED == 1
    return tDepth.Load( int2(IN.pos.xy)*2, 0 );
#else
    return tDepth.Load( int2(IN.pos.xy), 0 );
#endif

#endif
}

//----------------------------------------------------------------------------------
float LinearizeDepth_PS( PostProc_VSOut IN ) : SV_TARGET
{
#if NUM_MSAA_SAMPLES == 1
    float z = tDepth.SampleLevel( samNearest, IN.tex, 0 ).x;
    return g_ZFar*g_ZNear / (g_ZFar - z*(g_ZFar-g_ZNear));
#else
    return 0;
#endif
}

//-------------------------------------------------------------------------
technique10 ResolveND
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, ResolveND_PS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepth, 0 );
    }
}

technique10 ResolveD
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, ResolveD_PS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepth, 0 );
    }
}

technique10 LinearizeDepth
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, LinearizeDepth_PS() ) );
        SetBlendState( DisableBlend, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState( DisableDepth, 0 );
    }
}
