//----------------------------------------------------------------------------------
// File:   Scene3D.h
// Author: Miguel Sainz
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------
#include "Scene3D.h"

#define MAX_PATH_STR 512

// Vertex Data
struct Scene3DVertex
{
    D3DXVECTOR3 pos;
    D3DXVECTOR3 nor;
};

// Input layout definitions
D3D10_INPUT_ELEMENT_DESC VertexLayout[] =
{
    { "POSITION",  0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
    { "NORMAL",    0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
};

//-----------------------------------------------------------------------------
// Name: Scene3D
// Desc: Constructor
//-----------------------------------------------------------------------------
Scene3D::Scene3D()
{
    // Geometry
    m_VertexLayout   = NULL;
    m_VertexBuffer   = NULL;
    
    // Effects
    m_TechniqueSceneC   = NULL;
    m_TechniqueSceneCN  = NULL;
    m_TechniqueSceneCD  = NULL;
    m_TechniqueSceneCND = NULL;

    m_Effect         = NULL;
    m_VarWVP         = NULL;
    m_VarWV          = NULL;
    m_VarEye         = NULL;
}

Scene3D::~Scene3D()
{
  
}

//-----------------------------------------------------------------------------
// Name: OnCreateDevice
// Desc: 
//-----------------------------------------------------------------------------
HRESULT Scene3D::OnCreateDevice( ID3D10Device* pd3dDevice, MeshDescriptor *meshDesc )
{
    HRESULT hr = S_OK;

    m_D3DDevice = pd3dDevice;
    
    // Create the effect
    WCHAR str[MAX_PATH_STR+1];
    
    DXUTFindDXSDKMediaFileCch( str, MAX_PATH_STR, DEFAULT_SHADERPATH L"\\ScreenSpaceAO\\Scene3D.fx" );
    hr |= D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, m_D3DDevice, NULL, NULL, &m_Effect, NULL, &hr);
    V( m_Effect && m_Effect->IsValid() );

    // Obtain the technique
    m_TechniqueSceneC   = m_Effect->GetTechniqueByName( "RenderSceneDiffuse"            );
    m_TechniqueSceneCN  = m_Effect->GetTechniqueByName( "RenderSceneDiffuseNormal"      );
    m_TechniqueSceneCD  = m_Effect->GetTechniqueByName( "RenderSceneDiffuseDepth"       );
    m_TechniqueSceneCND = m_Effect->GetTechniqueByName( "RenderSceneDiffuseNormalDepth" );

    // Obtain pointers to the shader variales
    m_VarWVP           = m_Effect->GetVariableByName( "g_WorldViewProjection" )->AsMatrix();
    m_VarWV            = m_Effect->GetVariableByName( "g_WorldView" )->AsMatrix();
    m_VarWVIT          = m_Effect->GetVariableByName( "g_WorldViewIT" )->AsMatrix();
    m_VarEye           = m_Effect->GetVariableByName( "g_EyePosition" )->AsVector();
    m_VarIsGroundPlane = m_Effect->GetVariableByName( "g_isGroundPlane" )->AsScalar();

    // Create the input layout
    D3D10_PASS_DESC PassDesc;
    UINT numElements;

    m_TechniqueSceneC->GetPassByIndex( 0 )->GetDesc( &PassDesc );
    
    numElements = sizeof(VertexLayout)/sizeof(VertexLayout[0]);
    hr |= m_D3DDevice->CreateInputLayout( VertexLayout, numElements, PassDesc.pIAInputSignature, PassDesc.IAInputSignatureSize, &m_VertexLayout );
    
    // Load the meshes
    for (int i = 0; i < NUM_MESHES; i++){
        DXUTFindDXSDKMediaFileCch( str, MAX_PATH_STR, (LPCWSTR)meshDesc[i].path );
        hr |= m_Mesh[i].Create( m_D3DDevice, str );
        m_useGroundPlane[i] = meshDesc[i].useGroundPlane;
    }

    if( FAILED( hr ) ){ 
        MessageBox(NULL, L"Unable to create mesh!!!", L"ERROR", MB_OK|MB_SETFOREGROUND|MB_TOPMOST);		
        return FALSE;
    }

    for (int i = 0; i < NUM_MESHES; i++){
        m_bbExtents[i] = m_Mesh[i].GetMeshBBoxExtents(0);
        m_bbCenter[i]  = m_Mesh[i].GetMeshBBoxCenter(0);
    }

    // Create the floor plane
    float w = 1.0f;
    Scene3DVertex vertices[] =
    {
        { D3DXVECTOR3( -w, 0,  w ), D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) },
        { D3DXVECTOR3(  w, 0,  w ), D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) },
        { D3DXVECTOR3(  w, 0, -w ), D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) },
        { D3DXVECTOR3( -w, 0, -w ), D3DXVECTOR3( 0.0f, 1.0f, 0.0f ) },

    };
    D3D10_BUFFER_DESC bd;
    bd.Usage = D3D10_USAGE_DEFAULT;
    bd.ByteWidth = sizeof( Scene3DVertex ) * 4;
    bd.BindFlags = D3D10_BIND_VERTEX_BUFFER;
    bd.CPUAccessFlags = 0;
    bd.MiscFlags = 0;
    D3D10_SUBRESOURCE_DATA InitData;
    InitData.pSysMem = vertices;
    hr = pd3dDevice->CreateBuffer( &bd, &InitData, &m_VertexBuffer );
    if( FAILED(hr) )
        return hr;

    // Create index buffer
    DWORD indices[] =
    {
        0,1,2,
        0,2,3,
    };

    bd.Usage = D3D10_USAGE_DEFAULT;
    bd.ByteWidth = sizeof( DWORD ) * 6;
    bd.BindFlags = D3D10_BIND_INDEX_BUFFER;
    bd.CPUAccessFlags = 0;
    bd.MiscFlags = 0;
    InitData.pSysMem = indices;
    hr = pd3dDevice->CreateBuffer( &bd, &InitData, &m_IndexBuffer );
    if( FAILED(hr) )
        return hr;



    return S_OK;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void Scene3D::DrawModel( int modelID )
{
    UINT Strides[1];
    UINT Offsets[1];
    ID3D10Buffer* pVB[1];
    pVB[0] = m_Mesh[modelID].GetVB10(0,0);
    Strides[0] = (UINT)m_Mesh[modelID].GetVertexStride(0,0);
    Offsets[0] = 0;
    m_D3DDevice->IASetVertexBuffers( 0, 1, pVB, Strides, Offsets );
    m_D3DDevice->IASetIndexBuffer( m_Mesh[modelID].GetIB10(0), m_Mesh[modelID].GetIBFormat10(0), 0 );
    m_D3DDevice->IASetInputLayout( m_VertexLayout );

    for( UINT subset = 0; subset < m_Mesh[modelID].GetNumSubsets(0); ++subset )
    {
        SDKMESH_SUBSET* pSubset = m_Mesh[modelID].GetSubset( 0, subset );
        D3D10_PRIMITIVE_TOPOLOGY PrimType = m_Mesh[modelID].GetPrimitiveType10( (SDKMESH_PRIMITIVE_TYPE)pSubset->PrimitiveType );
        m_D3DDevice->IASetPrimitiveTopology( PrimType );
        m_Technique->GetPassByIndex( 0 )->Apply( 0 );
        m_D3DDevice->DrawIndexed( (UINT)pSubset->IndexCount, 0, (UINT)pSubset->VertexStart );
    }
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void Scene3D::DrawGroundPlane()
{
    UINT stride = sizeof( Scene3DVertex );
    UINT offset = 0;
    m_D3DDevice->IASetVertexBuffers    ( 0, 1, &m_VertexBuffer, &stride, &offset );
    m_D3DDevice->IASetIndexBuffer      ( m_IndexBuffer, DXGI_FORMAT_R32_UINT, 0 );
    m_D3DDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST );
    m_D3DDevice->IASetInputLayout      ( m_VertexLayout );

    m_Technique->GetPassByIndex( 0 )->Apply( 0 );
    m_D3DDevice->DrawIndexed( 6, 0, 0 );
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void Scene3D::UpdateMatrices( const D3DXMATRIX &mWorld,
                              const D3DXMATRIX &mView, 
                              const D3DXMATRIX &mProj )
{
    D3DXMATRIX WVPMatrix = mWorld * mView * mProj;
    D3DXMATRIX WVMatrix  = mWorld * mView;

    D3DXMATRIX WVMatrixI, WVMatrixIT;
    D3DXMatrixInverse(&WVMatrixI, NULL, &WVMatrix);
    D3DXMatrixTranspose(&WVMatrixIT, &WVMatrixI);

    m_VarWVIT->SetMatrix ( WVMatrixIT );
    m_VarWVP->SetMatrix  ( WVPMatrix );
    m_VarWV ->SetMatrix  ( WVMatrix );
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
HRESULT Scene3D::OnFrameRender( const D3DXMATRIX *p_mWorld,
                                const D3DXMATRIX *p_mView, 
                                const D3DXMATRIX *p_mProj,
                                int modelID )
{
    m_Technique = m_TechniqueSceneCND;
    
    UpdateMatrices( *p_mWorld, *p_mView, *p_mProj );

    m_VarIsGroundPlane->SetBool(false);

    DrawModel(modelID);

    if( m_useGroundPlane[modelID] )
    {
        D3DXMATRIX mTranslate;
        float h = m_bbCenter[modelID].y - m_bbExtents[modelID].y;
        D3DXMatrixTranslation( &mTranslate, 0.0f, h, 0.0f );
        D3DXMATRIX World = mTranslate * (*p_mWorld);

        UpdateMatrices( World, *p_mView, *p_mProj );

        m_VarIsGroundPlane->SetBool(true);

        DrawGroundPlane();
    }

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: OnDestroyDevice
//-----------------------------------------------------------------------------
HRESULT Scene3D::OnDestroyDevice()
{
    SAFE_RELEASE( m_Effect );
    SAFE_RELEASE( m_VertexLayout );

    for (int i = 0; i < NUM_MESHES; i++)
        m_Mesh[i].Destroy();

    SAFE_RELEASE( m_VertexBuffer );
    SAFE_RELEASE( m_IndexBuffer );

    return S_OK;
}

