//--------------------------------------------------------------------------------------
// File:    Scene3D.fx
// Authors: Miguel Sainz
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// PARAMETERS
//-----------------------------------------------------------------------------

matrix g_WorldViewProjection;
matrix g_WorldView;
matrix g_WorldViewIT;
float3 g_EyePosition;

float  g_PointLightIntensity = 1.0;
float  g_Kd = 1;
float  g_Ks = 2;
float  g_specPower = 200;
float4 g_diffuseColor  = float4(1.0,1.0,1.0,1.0);
float4 g_specularColor = float4(1.0,1.0,1.0,1.0);
bool   g_isGroundPlane = false;

//-----------------------------------------------------------------------------
// Shaders IN/OUT formats
//-----------------------------------------------------------------------------

struct VS_INPUT
{
    float3 Pos  : POSITION;
    float3 Norm : NORMAL;
};

struct VS_OUTPUT
{
    float4 HPosition    : SV_POSITION;
    float3 Normal       : NORMAL;
    float3 EyePos       : TEXCOORD0;
    float3 KeyLightPos  : TEXCOORD1;
};

struct PS_INPUT
{
    float4 HPosition    : SV_POSITION;
    float3 Normal       : NORMAL;
    float3 EyePos       : TEXCOORD0;
    float3 KeyLightPos  : TEXCOORD1;
    bool FrontFacing    : SV_IsFrontFace;
};

struct PS_OUTPUT
{
    float4 Color    : SV_Target0;
};

struct PS_OUTPUT_N
{
    float4 Color    : SV_Target0;
    float4 Normal   : SV_Target1;
};

struct PS_OUTPUT_ND
{
    float4 Color    : SV_Target0;
    float4 Normal   : SV_Target1;
    float LinDepth  : SV_Target2;
};

struct PS_OUTPUT_D
{
    float4 Color    : SV_Target0;
    float LinDepth  : SV_Target1;
};

BlendState NoBlending
{
    BlendEnable[0] = FALSE;
    BlendEnable[1] = FALSE;
    BlendEnable[2] = FALSE;
};

DepthStencilState DepthState
{
    DepthEnable = True;
    DepthFunc = Less;
};

RasterizerState MS_RS
{
    CullMode = None; // for the cathedral model
    MultisampleEnable = TRUE;
};

//-----------------------------------------------------------------------------
// Vertex Shader
//-----------------------------------------------------------------------------

VS_OUTPUT VS_Scene( VS_INPUT input )
{
    VS_OUTPUT output;
    
    output.HPosition = mul( float4(input.Pos,1), g_WorldViewProjection );
    
    output.EyePos = mul( float4(input.Pos,1), g_WorldView );
    
    output.Normal   = normalize( mul( input.Norm, (float3x3)g_WorldViewIT ) );
    
    // Transform the global lights to eye space
    output.KeyLightPos  = mul( float4( 0.0, 1.0, 0.0, 1.0), g_WorldView );
     
    return output;
}

//-----------------------------------------------------------------------------
// Pixel Shader
//-----------------------------------------------------------------------------

float4 calculateColor( PS_INPUT In ){

    if(g_isGroundPlane) return float4(1,1,1,1);

    if(!In.FrontFacing) In.Normal = -In.Normal;

    // Key light
    float4 KeylightColor = float4(.457,.722 , 0.0 ,1);
    float3 EyeToKeyLight = normalize( In.KeyLightPos  - In.EyePos );
    float4 diffuseLightingKeyLight = KeylightColor * (0.5 + saturate( dot(In.Normal, EyeToKeyLight)) * 0.5) ;

    // Fresnel term 
    float3 V = -normalize( In.EyePos );
    float fresnel = 0.1 + 0.9*pow(1.0 - max(0, dot(V, In.Normal)), 2.0);

    return ( 0.7 * diffuseLightingKeyLight + g_specularColor * 0.3 * fresnel);
}


PS_OUTPUT PS_Scene( PS_INPUT In )
{   
    PS_OUTPUT Out = (PS_OUTPUT)0;
   
    Out.Color = calculateColor( In );

    return Out;
}

PS_OUTPUT_N PS_SceneN( PS_INPUT In )
{   
    PS_OUTPUT_N Out = (PS_OUTPUT_N)0;
    
    Out.Color = calculateColor( In );

    // Use face normals instead of interpolated normals to make sure that the tangent planes
    // associated with the normals are actually tangent to the surface
    Out.Normal.xyz = normalize(cross(ddx(In.EyePos.xyz), ddy(In.EyePos.xyz)));

    return Out;
}

PS_OUTPUT_ND PS_SceneND( PS_INPUT In )
{   
    PS_OUTPUT_ND Out = (PS_OUTPUT_ND)0;
    
    Out.Color = calculateColor( In );

    Out.Normal.xyz = normalize(cross(ddx(In.EyePos.xyz), ddy(In.EyePos.xyz)));

    Out.LinDepth = In.EyePos.z;
    
    return Out;
}

PS_OUTPUT_D PS_SceneD( PS_INPUT In )
{   
    PS_OUTPUT_D Out = (PS_OUTPUT_D)0;
    
    Out.Color = calculateColor( In );

    Out.LinDepth = In.EyePos.z;
    
    return Out;
}

//-----------------------------------------------------------------------------
// TECHNIQUES
//-----------------------------------------------------------------------------
technique10 RenderSceneDiffuse
{
    pass p0
    {
        SetVertexShader  ( CompileShader( vs_4_0, VS_Scene( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, PS_Scene( ) ) );
        SetRasterizerState( MS_RS );
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState(DepthState, 0);
    }
}

technique10 RenderSceneDiffuseNormal
{
    pass p0
    {
        SetVertexShader  ( CompileShader( vs_4_0, VS_Scene( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, PS_SceneN( ) ) );
        SetRasterizerState( MS_RS );
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState(DepthState, 0);
    }
}

technique10 RenderSceneDiffuseNormalDepth
{
    pass p0
    {
        SetVertexShader  ( CompileShader( vs_4_0, VS_Scene( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, PS_SceneND( ) ) );
        SetRasterizerState( MS_RS );
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState(DepthState, 0);    
    }
}

technique10 RenderSceneDiffuseDepth
{
    pass p0
    {
        SetVertexShader  ( CompileShader( vs_4_0, VS_Scene( ) ) );
        SetGeometryShader( NULL );
        SetPixelShader   ( CompileShader( ps_4_0, PS_SceneD( ) ) );
        SetRasterizerState( MS_RS );
        SetBlendState( NoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetDepthStencilState(DepthState, 0);    
    }
}
