//----------------------------------------------------------------------------------
// File:   ScreenSpaceAO.cpp
// Author: Miguel Sainz & Louis Bavoil
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2008 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

#include "ScreenSpaceAO.h"
#include "SimpleRT.h"

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
CModelViewerCamera            g_UserCamera;             // A model viewing camera
D3DXVECTOR3                   g_ModelOffset;

CDXUTDialogResourceManager    g_DialogResourceManager;  // manager for shared resources of dialogs
CD3DSettingsDlg               g_SettingsDlg;            // Device settings dialog
CDXUTDialog                   g_HUD;                    // dialog for standard controls

bool                          g_ShowControls   = true;   // If true, it renders the UI control text
bool                          g_DrawUI         = true;
bool                          g_ShowAO         = true;
bool                          g_ShowDiffuse    = true;
bool                          g_UseHalfResAO   = true;
float                         g_fFrameTime     = 0.0f;
bool                          g_UseFullResDepthBlur = true;

int                           g_BBufferWidth;
int                           g_BBufferHeight;
int                           g_AOBufferWidth;
int                           g_AOBufferHeight;
int                           g_NSamples;

float                         g_fFovy;       
float                         g_fAspectRatio;
float                         g_zNear;      
float                         g_zFar;        

ID3DX10Font*                  g_pFont10        = NULL;   // Font for drawing text
ID3DX10Sprite*                g_pSprite10      = NULL;   // Sprite for batching draw text calls
CDXUTTextHelper*              g_pTxtHelper     = NULL;

ID3D10EffectMatrixVariable*   g_pmWorldViewProj;
ID3D10EffectMatrixVariable*   g_pmWorld;
ID3D10EffectScalarVariable*   g_pfTime;
ID3D10Device*                 g_pD3DDevice;

D3D10_VIEWPORT                g_Viewport;
D3D10_VIEWPORT                g_ViewportSSAO;


BilateralBlurEngine           g_BlurEngine;
ResolveEngine                 g_ResolveEngine;

Scene3D*                      g_pScene                 = NULL;
HorizonBasedAOEngine*	      g_HorizonBasedAOEngine   = NULL;
RayMarchingAOEngine*	      g_RayMarchingAOEngine    = NULL;

ScreenSpaceAOEngine*          g_currentSSAO            = NULL;
ScreenSpaceAOEngine*          g_SSAOEngineList[MAX_SSAO_ENGINES];

CDXUTDialog                   g_TechniqueUI   [MAX_SSAO_ENGINES];
CDXUTDialog                   g_BlurUI;
int                           g_CurrentMeshID = 0;

int g_SSAOEngineCount         =  0;  
int	g_SSAOCurrentEngine       = -1;
int g_MSAACurrentSettings     =  0;

#define WORLD_OFFSET 0.01f

//--------------------------------------------------------------------------------------

SimpleRT*                 g_ColorRT;
SimpleRT*                 g_NormalRT;
SimpleRT*                 g_DepthRT;

ID3D10DepthStencilView*             g_DSV;
ID3D10EffectShaderResourceVariable* g_ColorSRV;
ID3D10EffectShaderResourceVariable* g_NormalDSRV;
ID3D10Texture2D*                    g_pDepthStencil;

bool    g_DumpFrame;


WCHAR* g_szTechniques[] =
{
    L"Horizon Based",
    L"Ray Marching",
};

MeshDescriptor g_meshDesc[NUM_MESHES] =
{
    {L"Dragon",      L"..\\Media\\dragon.sdkmesh",      true,  0.35f },
    {L"Cornell Box", L"..\\Media\\cornell_box.sdkmesh", false, 0.35f },
    {L"Cathedral",   L"..\\Media\\sibenik.sdkmesh",     false, 0.01f },
};

struct MSAADescriptor
{
    WCHAR name[32];
    int samples;
    int quality;
};

const int NUM_MSAA_MODES = 7;

MSAADescriptor g_MSAADesc[NUM_MSAA_MODES] =
{
    {L"1x MSAA",   1, 0},
    {L"2x MSAA",   2, 0},
    {L"4x MSAA",   4, 0},
    {L"8x MSAA",   4, 8},
    {L"8xQ MSAA",  8, 8},
    {L"16x MSAA",  4, 16},
    {L"16xQ MSAA", 8, 16},
};

// Auxiliary functions (coded at the end)
void    InitApp();
HRESULT AllocateAllSurfaces();
void    RenderText();
void    RenderText10( double ftime );

//--------------------------------------------------------------------------------------
void DumpFrameNonMSAA(int frameId)
{
    IDXGISwapChain* pSwapChain = DXUTGetDXGISwapChain();
    ID3D10Texture2D* pRT;
    pSwapChain->GetBuffer(0, __uuidof(pRT), reinterpret_cast<void**>(&pRT));
    WCHAR filename[100];
    StringCchPrintf(filename, 100, L"P:\\screenshot%.3d.png", frameId); 
    D3DX10SaveTextureToFile(pRT, D3DX10_IFF_PNG, filename);
    pRT->Release();
}

//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, 
                                       DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    // Disable VSync
    pDeviceSettings->d3d10.SyncInterval = 0;

    // Disable MSAA
    pDeviceSettings->d3d10.sd.SampleDesc.Count = 1;
    pDeviceSettings->d3d10.sd.SampleDesc.Quality = 0;

    // Do not allocate a depth buffer
    pDeviceSettings->d3d10.AutoCreateDepthStencil = 0;

    return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    g_pD3DDevice = pd3dDevice;

    V_RETURN( g_DialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( CDXUTDirectionWidget::StaticOnD3D10CreateDevice( pd3dDevice ) );
   
    // Initialize the font
    V_RETURN( D3DX10CreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, 
                                OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
                                L"Arial", &g_pFont10 ) );

    V_RETURN( D3DX10CreateSprite( pd3dDevice, 512, &g_pSprite10 ) );
    g_pTxtHelper = new CDXUTTextHelper( NULL, NULL, g_pFont10, g_pSprite10, 15 );


    D3DXVECTOR3 vecEye(0.0f, 2.0f, 0.0f);
    D3DXVECTOR3 vecAt (0.0f, 0.0f, 0.0f);
    g_UserCamera.SetViewParams( &vecEye, &vecAt );

    g_UserCamera.SetRadius(1.5f, 0.01f);

  
    // Load the Models
    g_pScene->OnCreateDevice( pd3dDevice, g_meshDesc );

    for(int i = 0; i < MAX_SSAO_ENGINES; i++) {
        g_SSAOEngineList[i]->OnCreateDevice(pd3dDevice);
    }

    g_currentSSAO = g_SSAOEngineList[g_SSAOCurrentEngine];

    g_BlurEngine.OnCreateDevice( pd3dDevice );
    g_ResolveEngine.OnCreateDevice( pd3dDevice );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    // SwapChain has changed and may have new attributes such as size. Application
    // should create resources dependent on SwapChain, such as 2D textures that
    // need to have the same size as the back buffer.
    DXUTTRACE( L"SwapChain Reset called\n" );

    V_RETURN( g_DialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
    V_RETURN( g_SettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

    g_pD3DDevice     = pd3dDevice;

    g_BBufferWidth   = pBackBufferSurfaceDesc->Width ;
    g_BBufferHeight  = pBackBufferSurfaceDesc->Height;
    
    if (g_UseHalfResAO) {
        g_AOBufferWidth  = g_BBufferWidth / 2;
        g_AOBufferHeight = g_BBufferHeight / 2;
    } else {
        g_AOBufferWidth  = g_BBufferWidth;
        g_AOBufferHeight = g_BBufferHeight;
    }

    // Create a viewport to match the screen size
    g_Viewport.TopLeftX = 0;
    g_Viewport.TopLeftY = 0;
    g_Viewport.MinDepth = 0;
    g_Viewport.MaxDepth = 1;
    g_Viewport.Width    = g_BBufferWidth;
    g_Viewport.Height   = g_BBufferHeight;

    // Create a viewport for SSAO pass
    g_ViewportSSAO.TopLeftX = 0;
    g_ViewportSSAO.TopLeftY = 0;
    g_ViewportSSAO.MinDepth = 0;
    g_ViewportSSAO.MaxDepth = 1;
    g_ViewportSSAO.Width    = g_AOBufferWidth;
    g_ViewportSSAO.Height   = g_AOBufferHeight;

    // Setup the camera's projection parameters
    g_fFovy        = 40 * D3DX_PI / 180;
    g_fAspectRatio = g_BBufferWidth / (float)g_BBufferHeight;
    g_zNear        = 0.01f;
    g_zFar         = 500.0f;

    g_UserCamera.SetProjParams ( g_fFovy, g_fAspectRatio, g_zNear, g_zFar );
    g_UserCamera.SetWindow     ( g_BBufferWidth, g_BBufferHeight );
    g_UserCamera.SetButtonMasks( MOUSE_LEFT_BUTTON, MOUSE_WHEEL, 0 );
    
    g_HUD.SetLocation( g_BBufferWidth - 170, 0 );
    g_HUD.SetSize    ( 170, 170 );

    for(int i = 0; i < MAX_SSAO_ENGINES; i++){
        g_TechniqueUI[i].SetLocation( g_BBufferWidth - 170, 230 );
        g_TechniqueUI[i].SetSize( 170, 170 );
    }
    g_BlurUI.SetLocation( g_BBufferWidth - 170, 300 );
    g_BlurUI.SetSize( 170, 170 );

    // Allocate all rendering surfaces, as the SwapChain only contains a non MSAA color buffer.
    hr = AllocateAllSurfaces();

    return hr;
}


//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{	
    g_UserCamera.FrameMove(fElapsedTime);
}


//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    ID3D10RenderTargetView* pBBufferRTV = DXUTGetD3D10RenderTargetView();

    int activeRTs = 3;
    ID3D10RenderTargetView* MRTs[3];
    MRTs[0] = g_ColorRT->pRTV;
    MRTs[1] = g_NormalRT->pRTV;
    MRTs[2] = g_DepthRT->pRTV;

    // If the settings dialog is being shown, then render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }

    float White[4] = { 1.0f, 1.0f, 1.0f, 0.0f };
    pd3dDevice->ClearRenderTargetView( g_ColorRT->pRTV, White );

    float MaxDepth[4] = { 1000.0f };
    pd3dDevice->ClearRenderTargetView( g_DepthRT->pRTV, MaxDepth );

    pd3dDevice->ClearDepthStencilView( g_DSV, D3D10_CLEAR_DEPTH, 1.0, 0 );

    pd3dDevice->RSSetViewports(1, &g_Viewport );
    
    pd3dDevice->OMSetRenderTargets( activeRTs, MRTs, g_DSV );

    D3DXMATRIX TranslationMatrix;
    D3DXMatrixTranslation( &TranslationMatrix, g_ModelOffset[0], g_ModelOffset[1], g_ModelOffset[2]);
    D3DXMATRIX WorldMatrix = TranslationMatrix * (*g_UserCamera.GetWorldMatrix());

    g_pScene->OnFrameRender ( &WorldMatrix, 
                              g_UserCamera.GetViewMatrix(), 
                              g_UserCamera.GetProjMatrix(), 
                              g_CurrentMeshID );

    // Resolve color buffer
    ID3D10ShaderResourceView* pColorSRV = NULL;
    g_ResolveEngine.ResolveColors(g_ColorRT, &pColorSRV);

    // Resolve normals and depths
    ID3D10ShaderResourceView* pNormalSRV = NULL;
    ID3D10ShaderResourceView* pDepthSRV = NULL;
    ID3D10ShaderResourceView* pFullResDepthSRV = NULL;
    g_ResolveEngine.ResolveNormalsAndDepths(g_NormalRT->pSRV,
                                            g_DepthRT->pSRV,
                                            NULL,
                                            &pNormalSRV,
                                            &pDepthSRV, 
                                            &pFullResDepthSRV);

    if( !g_ShowDiffuse )
        pColorSRV = NULL;

    // Run SSAO path
    if (g_ShowAO) {
        g_currentSSAO->OnFrameRender( pDepthSRV, pNormalSRV );
        
        // Run the blur passes
        ID3D10ShaderResourceView* MSAADepthSRV = g_DepthRT->pSRV;
        g_BlurEngine.OnFrameRender( pBBufferRTV, 
                                    pColorSRV, 
                                    g_currentSSAO->getAmbientOcclusionRT()->pSRV,
                                    pDepthSRV, 
                                    g_UseFullResDepthBlur ? pFullResDepthSRV : pDepthSRV,
                                    MSAADepthSRV);
        
    } else {
        // Run the pass-through pass
        g_BlurEngine.OnFrameRender( pBBufferRTV, pColorSRV );
    }
    
    // Render the HUD
    if (g_DrawUI) {
        RenderText10( fElapsedTime );
        if( g_ShowControls )
        {
            g_TechniqueUI[g_SSAOCurrentEngine].OnRender( fElapsedTime );
            g_BlurUI.OnRender( fElapsedTime );
        }
        g_HUD.OnRender( fElapsedTime ); 
    }

    static int g_Frame = 0;
    if (g_DumpFrame) {
        DumpFrameNonMSAA(g_Frame);
        ++g_Frame;
    }
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext )
{
    DXUTTRACE( L"SwapChain Lost called\n" );

    g_DialogResourceManager.OnD3D10ReleasingSwapChain();

    for(int i = 0; i < MAX_SSAO_ENGINES; i++)
        g_SSAOEngineList[i]->OnReleasingSwapChain();

    g_BlurEngine.OnReleasingSwapChain();
    g_ResolveEngine.OnReleasingSwapChain();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10DestroyDevice();
    g_SettingsDlg.OnD3D10DestroyDevice();
    CDXUTDirectionWidget::StaticOnD3D10DestroyDevice();
    
    SAFE_RELEASE( g_pFont10    );
    SAFE_RELEASE( g_pSprite10  );
    SAFE_DELETE ( g_pTxtHelper );
    g_pScene->OnDestroyDevice();

    for(int i = 0; i < MAX_SSAO_ENGINES; i++)
        g_SSAOEngineList[i]->OnDestroyDevice();

    g_BlurEngine.OnDestroyDevice();
    g_ResolveEngine.OnDestroyDevice();

    SAFE_RELEASE(g_pDepthStencil);
    SAFE_RELEASE(g_DSV);

    SAFE_DELETE(g_ColorRT);
    SAFE_DELETE(g_NormalRT);
    SAFE_DELETE(g_DepthRT);

    SAFE_DELETE( g_HorizonBasedAOEngine );
    SAFE_DELETE( g_RayMarchingAOEngine );
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, 
                          bool* pbNoFurtherProcessing, void* pUserContext )
{
    // Always allow dialog resource manager calls to handle global messages
    // so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    *pbNoFurtherProcessing = g_TechniqueUI[g_SSAOCurrentEngine].MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    *pbNoFurtherProcessing = g_BlurUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    
    // Pass all windows messages to camera so it can respond to user input
    g_UserCamera.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}

//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN:   DXUTToggleFullScreen(); break;
        case IDC_TOGGLEREF:          DXUTToggleREF();        break;
           
        case IDC_CHANGEDEVICE:       
            g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); 
            break;
        case IDC_CHANGETECHNIQUE:
        {
            CDXUTComboBox* pComboBox = (CDXUTComboBox*) pControl;
            g_SSAOCurrentEngine = pComboBox->GetSelectedIndex();
            g_currentSSAO = g_SSAOEngineList[g_SSAOCurrentEngine];
            g_currentSSAO->SetSceneScale( g_meshDesc[ g_CurrentMeshID ].scale );
            g_BlurEngine.SetSceneScale( g_meshDesc[ g_CurrentMeshID ].scale );
            break;
        }
        case IDC_CHANGEMESH:
        {
            CDXUTComboBox* pComboBox = (CDXUTComboBox*) pControl;
            g_CurrentMeshID = pComboBox->GetSelectedIndex();
            g_currentSSAO->SetSceneScale( g_meshDesc[ g_CurrentMeshID ].scale );
            g_BlurEngine.SetSceneScale( g_meshDesc[ g_CurrentMeshID ].scale );
            break;
        }
        case IDC_CHANGEMSAA:
        {
            CDXUTComboBox* pComboBox = (CDXUTComboBox*) pControl;
            g_MSAACurrentSettings = pComboBox->GetSelectedIndex();
            AllocateAllSurfaces();
            break;
        }
        case IDC_SHOW_AO:
        {
            g_ShowAO = g_HUD.GetCheckBox(IDC_SHOW_AO)->GetChecked();
            break;
        }
        case IDC_SHOW_DIFFUSE:
        {
            g_ShowDiffuse = g_HUD.GetCheckBox(IDC_SHOW_DIFFUSE)->GetChecked();
            break;
        }
        case IDC_USE_HALF_RES_AO:
        {
            g_UseHalfResAO = g_HUD.GetCheckBox(IDC_USE_HALF_RES_AO)->GetChecked();
            
            if (g_UseHalfResAO)
            {
                g_AOBufferWidth  = g_BBufferWidth / 2;
                g_AOBufferHeight = g_BBufferHeight / 2;
            }
            else
            {
                g_AOBufferWidth  = g_BBufferWidth;
                g_AOBufferHeight = g_BBufferHeight;
            }
            
            // Modify the viewport
            g_ViewportSSAO.Width    = g_AOBufferWidth;
            g_ViewportSSAO.Height   = g_AOBufferHeight;

            // Update the engines
            for(int i = 0; i < MAX_SSAO_ENGINES; i++)
                g_SSAOEngineList[i]->OnResizedSwapChain( g_AOBufferWidth,
                                                         g_AOBufferHeight,
                                                         g_fFovy );

            g_BlurEngine.OnResizedSwapChain( g_AOBufferWidth, 
                                             g_AOBufferHeight,
                                             g_BBufferWidth, 
                                             g_BBufferHeight );

            g_ResolveEngine.OnResizedSwapChain( g_BBufferWidth,
                                                g_BBufferHeight,
                                                g_NSamples,
                                                g_AOBufferWidth,
                                                g_AOBufferHeight,
                                                g_zFar, g_zNear );
            break;
        }
    }
}


//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnTechniqueGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    g_currentSSAO->OnGUIEvent( &g_TechniqueUI[g_SSAOCurrentEngine], nEvent, nControlID, pControl, pUserContext );
    g_BlurEngine.OnGUIEvent( &g_BlurUI, nEvent, nControlID, pControl, pUserContext );
}

//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
    if( bKeyDown )
    {
        switch( nChar )
        {
            case 'C': g_ShowControls = !g_ShowControls; break;
            case 'U': g_DrawUI = !g_DrawUI; break;
            
            case 'P': g_ShowAO  = !g_ShowAO; break;

            case 'A': g_ModelOffset[0] -= WORLD_OFFSET; break;
            case 'D': g_ModelOffset[0] += WORLD_OFFSET; break;

            case 'W': g_ModelOffset[1] += WORLD_OFFSET; break;
            case 'S': g_ModelOffset[1] -= WORLD_OFFSET; break;

            case 'Q': g_ModelOffset[2] -= WORLD_OFFSET; break;
            case 'E': g_ModelOffset[2] += WORLD_OFFSET; break;

            case 'F': g_UseFullResDepthBlur = !g_UseFullResDepthBlur; break;

            case '1':
            case '2':
            case '3':
            case '4':
                g_SSAOCurrentEngine = nChar - '1';
                g_currentSSAO = g_SSAOEngineList[g_SSAOCurrentEngine];
                break;

            case 'R':
                g_currentSSAO->RecompileShader();
                g_BlurEngine.RecompileShader();
                g_ResolveEngine.RecompileShader();
                break;
        }
    }
}


//--------------------------------------------------------------------------------------
// Handle mouse button presses
//--------------------------------------------------------------------------------------
void CALLBACK OnMouse( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, 
                       bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, 
                       int xPos, int yPos, void* pUserContext )
{
}


//--------------------------------------------------------------------------------------
// Call if device was removed.  Return true to find a new device, false to quit
//--------------------------------------------------------------------------------------
bool CALLBACK OnDeviceRemoved( void* pUserContext )
{
    return true;
}

//--------------------------------------------------------------------------------------
// Initialize everything and go into a render loop
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // DXUT will create and use the best device (either D3D9 or D3D10) 
    // that is available on the system depending on which D3D callbacks are set below

    // Set general DXUT callbacks
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackMouse( OnMouse );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
    DXUTSetCallbackDeviceRemoved( OnDeviceRemoved );

    // Set the D3D10 DXUT callbacks. Remove these sets if the app doesn't need to support D3D10
    DXUTSetCallbackD3D10DeviceAcceptable  ( IsD3D10DeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated     ( OnD3D10CreateDevice );
    DXUTSetCallbackD3D10SwapChainResized  ( OnD3D10ResizedSwapChain );
    DXUTSetCallbackD3D10FrameRender       ( OnD3D10FrameRender );
    DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10ReleasingSwapChain );
    DXUTSetCallbackD3D10DeviceDestroyed   ( OnD3D10DestroyDevice );

    // Perform any application-level initialization here
    InitApp();

    DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
    DXUTCreateWindow( L"Screen Space Ambient Occlusion" );
    DXUTCreateDevice( true, 800, 800 );

    DXUTMainLoop(); // Enter into the DXUT render loop

    // Perform any application-level cleanup here

    return DXUTGetExitCode();
}


//--------------------------------------------------------------------------------------
// Initialize the app 
//--------------------------------------------------------------------------------------
void InitApp()
{
    // Initialize dialogs
    g_SettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init        ( &g_DialogResourceManager );

    g_HUD.SetCallback( OnGUIEvent ); int iY = 10; 
    g_HUD.AddButton  ( IDC_TOGGLEFULLSCREEN,   L"Toggle full screen" , 35, iY, 125, 22 );
    g_HUD.AddButton  ( IDC_TOGGLEREF,          L"Toggle REF (F3)"    , 35, iY += 24, 125, 22, VK_F3 );
    g_HUD.AddButton  ( IDC_CHANGEDEVICE,       L"Change device (F2)" , 35, iY += 24, 125, 22, VK_F2 );
    iY += 24;
    CDXUTComboBox *pComboBox;

    g_HUD.AddComboBox( IDC_CHANGEMSAA,    35, iY += 24, 125, 22, 'X', false, &pComboBox);
    for(char i = 0; i < NUM_MSAA_MODES; i++){
        pComboBox->AddItem(g_MSAADesc[i].name, (void*)i);
    }

    g_HUD.AddComboBox( IDC_CHANGETECHNIQUE,    35, iY += 24, 125, 22, 'T', false, &pComboBox);
    for(char i = 0; i < MAX_SSAO_ENGINES; i++){
        g_TechniqueUI[i].Init( &g_DialogResourceManager );
        g_TechniqueUI[i].SetCallback( OnGUIEvent );
        pComboBox->AddItem(g_szTechniques[i], (void*)i);
    }
    g_BlurUI.Init( &g_DialogResourceManager );
    g_BlurUI.SetCallback( OnGUIEvent );


    g_HUD.AddComboBox( IDC_CHANGEMESH,    35, iY += 24, 125, 22, 'M', false, &pComboBox);
    for(char i = 0; i < NUM_MESHES; i++){
        pComboBox->AddItem(g_meshDesc[i].name, (void*)i);
    }

    g_HUD.AddCheckBox( IDC_USE_HALF_RES_AO,  L"Half-Res AO",   35, iY += 34, 125, 22, g_UseHalfResAO );
    g_HUD.AddCheckBox( IDC_SHOW_DIFFUSE,  L"Show Diffuse" , 35, iY += 24, 125, 22, g_ShowDiffuse );
    g_HUD.AddCheckBox( IDC_SHOW_AO,       L"Show AO"      , 35, iY += 24, 125, 22, g_ShowAO      );
    
    // Objects for rendering
    g_pScene     = new Scene3D();
    
    // SSAO engines
    g_SSAOEngineCount   = 0;
    g_SSAOCurrentEngine = 0;

    // Add new engine
    g_HorizonBasedAOEngine = new HorizonBasedAOEngine();
    g_SSAOEngineList[g_SSAOEngineCount] = (ScreenSpaceAOEngine*)g_HorizonBasedAOEngine;
    g_TechniqueUI   [g_SSAOEngineCount].SetCallback( OnTechniqueGUIEvent );
    g_HorizonBasedAOEngine->CreateMenu( &g_TechniqueUI[g_SSAOEngineCount] );
    g_SSAOEngineCount++;

    g_RayMarchingAOEngine = new RayMarchingAOEngine();
    g_SSAOEngineList[g_SSAOEngineCount] = (ScreenSpaceAOEngine*)g_RayMarchingAOEngine;
    g_TechniqueUI   [g_SSAOEngineCount].SetCallback( OnTechniqueGUIEvent );
    g_RayMarchingAOEngine->CreateMenu( &g_TechniqueUI[g_SSAOEngineCount] );
    g_SSAOEngineCount++;

    for(int i = 0; i < MAX_SSAO_ENGINES; i++) {
        g_SSAOEngineList[i]->SetSceneScale( g_meshDesc[ g_CurrentMeshID ].scale );
    }

    g_BlurUI.SetCallback( OnTechniqueGUIEvent );
    g_BlurEngine.CreateMenu( &g_BlurUI );
    g_BlurEngine.SetSceneScale( g_meshDesc[ g_CurrentMeshID ].scale );
}

//--------------------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------------------
void CreateRenderTargets()
{
    D3D10_TEXTURE2D_DESC texDesc;
    texDesc.Width              = g_BBufferWidth;
    texDesc.Height             = g_BBufferHeight;
    texDesc.ArraySize          = 1;
    texDesc.MiscFlags          = 0;
    texDesc.MipLevels          = 1;
    texDesc.SampleDesc.Count   = g_MSAADesc[g_MSAACurrentSettings].samples;
    texDesc.SampleDesc.Quality = g_MSAADesc[g_MSAACurrentSettings].quality;

    texDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
    SAFE_DELETE(g_ColorRT);
    g_ColorRT = new SimpleRT( g_pD3DDevice, &texDesc );

    texDesc.Format = DXGI_FORMAT_R8G8B8A8_SNORM;
    SAFE_DELETE(g_NormalRT);
    g_NormalRT = new SimpleRT( g_pD3DDevice, &texDesc );

    texDesc.Format = DXGI_FORMAT_R32_FLOAT;
    SAFE_DELETE(g_DepthRT);
    g_DepthRT = new SimpleRT( g_pD3DDevice, &texDesc );
}

//--------------------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------------------
void CreateDepthBuffer()
{
    HRESULT hr;

    D3D10_TEXTURE2D_DESC texDesc;
    texDesc.ArraySize          = 1;
    texDesc.BindFlags          = D3D10_BIND_DEPTH_STENCIL;
    texDesc.CPUAccessFlags     = NULL;
    texDesc.Width              = g_BBufferWidth;
    texDesc.Height             = g_BBufferHeight;
    texDesc.MipLevels          = 1;
    texDesc.MiscFlags          = NULL;
    texDesc.SampleDesc.Count   = g_MSAADesc[g_MSAACurrentSettings].samples;
    texDesc.SampleDesc.Quality = g_MSAADesc[g_MSAACurrentSettings].quality;
    texDesc.Usage              = D3D10_USAGE_DEFAULT;
    texDesc.Format             = DXGI_FORMAT_R24G8_TYPELESS;

    SAFE_RELEASE(g_pDepthStencil);
    V( g_pD3DDevice->CreateTexture2D( &texDesc, NULL, &g_pDepthStencil) );
}

//--------------------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------------------
void CreateDepthStencilView()
{
    HRESULT hr;

    D3D10_DEPTH_STENCIL_VIEW_DESC dsvDesc;
    dsvDesc.Format        = DXGI_FORMAT_D24_UNORM_S8_UINT;
    if ( g_MSAADesc[g_MSAACurrentSettings].samples > 1 ) {
        dsvDesc.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2DMS;
    } else {
        dsvDesc.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2D;
    }
    dsvDesc.Texture2D.MipSlice = 0;

    SAFE_RELEASE(g_DSV);
    V( g_pD3DDevice->CreateDepthStencilView( g_pDepthStencil, &dsvDesc, &g_DSV) );
}

//--------------------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------------------
HRESULT AllocateAllSurfaces()
{
    g_NSamples = g_MSAADesc[g_MSAACurrentSettings].samples;

    g_BlurEngine.SetNumSamplesPerPixel(g_NSamples);

    CreateRenderTargets();

    CreateDepthBuffer();

    CreateDepthStencilView();

    // Update the engines
    for(int i = 0; i < MAX_SSAO_ENGINES; i++)
    {
        g_SSAOEngineList[i]->OnResizedSwapChain( g_AOBufferWidth, g_AOBufferHeight, g_fFovy );
    }

    g_BlurEngine.OnResizedSwapChain( g_AOBufferWidth, 
                                     g_AOBufferHeight,
                                     g_BBufferWidth, 
                                     g_BBufferHeight );

    g_ResolveEngine.OnResizedSwapChain( g_BBufferWidth,
                                        g_BBufferHeight,
                                        g_NSamples,
                                        g_AOBufferWidth,
                                        g_AOBufferHeight,
                                        g_zFar, g_zNear );

    return S_OK;
}

//--------------------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------------------
void RenderText()
{
    CDXUTTextHelper txtHelper(g_pFont10, NULL, 15);
    txtHelper.Begin();
    txtHelper.SetInsertionPos( 5, 5 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 0.0f, 0.0f, 1.0f ) );
    txtHelper.DrawTextLine( DXUTGetFrameStats(true) );
    txtHelper.DrawTextLine( DXUTGetDeviceStats() );
    txtHelper.End();
}

void RenderText10( double fTime )
{
    // Output statistics
    g_pTxtHelper->Begin();
    g_pTxtHelper->SetInsertionPos( 5, 5 );
    g_pTxtHelper->SetForegroundColor( D3DXCOLOR( 1.0f, 0.0f, 0.0f, 1.0f ) );
    g_pTxtHelper->DrawTextLine( DXUTGetFrameStats(true) );
    g_pTxtHelper->DrawTextLine( DXUTGetDeviceStats() );
    g_pTxtHelper->DrawFormattedTextLine( L"Time: %0.1f ms", (1.0 / DXUTGetFPS())*1000.0);

    g_pTxtHelper->End();
}
